import logging
import sys
from dataclasses import dataclass
from typing import Optional

from simple_parsing import Serializable


logger = logging.getLogger(__name__)


@dataclass
class TrainingArgs(Serializable):
    """The arguments for the trainer."""

    output_dir: str  # The output directory where the model predictions and checkpoints will be written.
    device: str = "gpu"  # The device to use for training. It can be "cpu" or "gpu".
    do_train: bool = True  # Whether to run training or not. This argument is not directly used by Trainer.
    do_eval: bool = False  # Whether to run the evaluation. This argument is not directly used by Trainer.
    do_predict: bool = False  # Whether to run the prediction. This argument is not directly used by Trainer.
    resume_from_checkpoint: str = "no"  # The checkpoint to resume from. It can be "no", "latest", "best" or a specific path. Tt will start from scratch by default.
    seed: int = 20220815  # The seed to use for reproducibility.

    dataloader_drop_last: bool = True
    dataloader_num_workers: int = 0
    dataloader_pin_memory: bool = True
    dataloader_persistent_workers: bool = False
    dataloader_prefetch_factor: Optional[int] = None
    per_device_train_batch_size: int = 8
    eval_batch_size: int = 1

    debug: bool = False
    num_train_epochs: int = 200  # The maximum number of epochs to run.
    max_steps: int = 0  # The maximum number of steps to run.
    max_grad_norm: float = 1.0  # The maximum gradient norm to clip.
    greater_is_better: bool = True  # Whether the metric is better when greater.
    save_epoch_interval: int = 1  # The interval to save the checkpoint.
    eval_epoch_interval: int = 1  # The interval to validate the model.
    gradient_accumulation_steps: int = 1
    ddp_find_unused_parameters: bool = False
    metric_for_best_model: str = ""  # The metric to use for saving the best model.
    early_stopping_patience: int = 20  # The patience for early stopping.
    save_total_limit: int = sys.maxsize  # The maximum number of checkpoints to keep.
    plot_lr: bool = False  # Whether to plot the learning rate.

    optim: str = "adamw"  # The optimizer to use.
    learning_rate: float = 1e-3  # The learning rate.
    adam_beta1: float = 0.9  # The beta1 for Adam.
    adam_beta2: float = 0.999
    adam_epsilon: float = 1e-8
    lr_scheduler_type: str = "constant_schedule_with_warmup"  # The type of learning rate scheduler.
    warmup_steps: int = 0  # Linear warmup over warmup_steps.
    warmup_ratio: float = 0.0  # Linear warmup over warmup_ratio * max_steps.
    cosine_with_min_lr_schedule_with_warmup_num_cycles: float = 0.5
    cosine_with_min_lr_schedule_with_warmup_initial_lr: float = 0.0
    cosine_with_min_lr_schedule_with_warmup_min_lr: Optional[float] = None

    acoustic_n_fft: int = 512  # The number of FFT points.
    acoustic_hop_length: int = 256
    acoustic_win_length: int = 512
    acoustic_sr: int = 16000  # The sample rate of the audio.



"""
contains building blocks used for the DNN-based estimators
mostly based on https://github.com/naplab/Conv-TasNet
"""

import numpy as np
import torch
from torch import nn

from audiozen.constant import EPSILON


class cLN(nn.Module):
    """Cumulative layer normalization."""

    def __init__(self, dimension, eps=EPSILON, trainable=True):
        super().__init__()

        self.eps = eps
        self.gain = nn.Parameter(torch.ones((1, dimension, 1), requires_grad=trainable))
        self.bias = nn.Parameter(torch.zeros((1, dimension, 1), requires_grad=trainable))

    def forward(self, input):
        # input size: (Batch, Freq, Time)
        # cumulative mean for each time step

        channel = input.size(1)
        time_step = input.size(2)

        step_sum = input.sum(1)  # B, T
        step_pow_sum = input.pow(2).sum(1)  # B, T
        cum_sum = torch.cumsum(step_sum, dim=1)  # B, T
        cum_pow_sum = torch.cumsum(step_pow_sum, dim=1)  # B, T

        entry_cnt = np.arange(channel, channel * (time_step + 1), channel)
        entry_cnt = torch.from_numpy(entry_cnt).type(input.type())
        entry_cnt = entry_cnt.view(1, -1).expand_as(cum_sum)

        cum_mean = cum_sum / entry_cnt  # B, T
        cum_var = (cum_pow_sum - 2 * cum_mean * cum_sum) / entry_cnt + cum_mean.pow(2)  # B, T
        cum_std = (cum_var + self.eps).sqrt()  # B, T

        cum_mean = cum_mean.unsqueeze(1)
        cum_std = cum_std.unsqueeze(1)

        x = (input - cum_mean.expand_as(input)) / cum_std.expand_as(input)
        return x * self.gain.expand_as(x).type(x.type()) + self.bias.expand_as(x).type(x.type())


class DepthConv1d(nn.Module):
    """Depthwise separable convolution."""

    def __init__(
        self,
        input_channel,
        hidden_channel,
        kernel,
        padding,
        dilation=1,
        skip=True,
        causal=False,
    ):
        super().__init__()

        self.causal = causal
        self.skip = skip

        self.conv1d = nn.Conv1d(input_channel, hidden_channel, 1)
        if self.causal:
            self.padding = (kernel - 1) * dilation
        else:
            self.padding = padding
        self.dconv1d = nn.Conv1d(
            hidden_channel,
            hidden_channel,
            kernel,
            dilation=dilation,
            groups=hidden_channel,
            padding=self.padding,
        )
        self.res_out = nn.Conv1d(hidden_channel, input_channel, 1)
        self.nonlinearity1 = nn.PReLU()
        self.nonlinearity2 = nn.PReLU()
        if self.causal:
            self.reg1 = cLN(hidden_channel, eps=EPSILON)
            self.reg2 = cLN(hidden_channel, eps=EPSILON)
        else:
            self.reg1 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)
            self.reg2 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)

        if self.skip:
            self.skip_out = nn.Conv1d(hidden_channel, input_channel, 1)

    def forward(self, input):
        output = self.reg1(self.nonlinearity1(self.conv1d(input)))
        if self.causal:
            output = self.reg2(self.nonlinearity2(self.dconv1d(output)[:, :, : -self.padding]))
        else:
            output = self.reg2(self.nonlinearity2(self.dconv1d(output)))
        residual = self.res_out(output)
        if self.skip:
            skip = self.skip_out(output)
            return residual, skip
        else:
            return residual


class TCNEstimator(nn.Module):
    """
    small modification of TCN for spectrum-based parameter estimation
    based on TCN implementation in https://github.com/naplab/Conv-TasNet
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        BN_dim,
        hidden_dim,
        layer=8,
        stack=3,
        kernel=3,
        skip=True,
        causal=True,
        dilated=True,
    ):
        super().__init__()

        # input is a sequence of features of shape (B, N, L)
        # here, N/2 refers to the number of frequencies, stacked real and imaginary components

        # normalization
        if not causal:
            self.LN = nn.GroupNorm(1, input_dim, eps=EPSILON)
        else:
            self.LN = cLN(input_dim, eps=EPSILON)

        self.BN = nn.Conv1d(input_dim, BN_dim, 1)

        # TCN for feature extraction
        self.receptive_field = 0
        self.dilated = dilated

        self.TCN = nn.ModuleList([])
        for s in range(stack):
            for i in range(layer):
                if self.dilated:
                    self.TCN.append(
                        DepthConv1d(
                            BN_dim,
                            hidden_dim,
                            kernel,
                            dilation=2**i,
                            padding=2**i,
                            skip=skip,
                            causal=causal,
                        )
                    )
                else:
                    self.TCN.append(
                        DepthConv1d(
                            BN_dim,
                            hidden_dim,
                            kernel,
                            dilation=1,
                            padding=1,
                            skip=skip,
                            causal=causal,
                        )
                    )
                if i == 0 and s == 0:
                    self.receptive_field += kernel
                else:
                    if self.dilated:
                        self.receptive_field += (kernel - 1) * 2**i
                    else:
                        self.receptive_field += kernel - 1

        # print("receptive field: {:d} time steps".format(self.receptive_field))

        self.output = nn.Conv1d(BN_dim, output_dim, 1)

        self.skip = skip

    def forward(self, input):
        # input shape: (B, N, L)
        # normalization
        output = self.BN(self.LN(input))

        # pass to TCN
        if self.skip:
            skip_connection = 0.0
            for i in range(len(self.TCN)):
                residual, skip = self.TCN[i](output)
                output = output + residual
                skip_connection = skip_connection + skip
        else:
            for i in range(len(self.TCN)):
                residual = self.TCN[i](output)
                output = output + residual

        # output layer
        if self.skip:
            output = self.output(skip_connection)
        else:
            output = self.output(output)

        return output



import importlib
from functools import partial
from typing import Iterable

import torch
import torch.nn as nn
import torch.nn.functional as F
from numpy import ndarray
from torch import Tensor
from torch.autograd import Function

from audiozen.acoustics.audio_feature import istft, stft


def numpy_to_torch(*args):
    out = []
    for arg in args:
        if isinstance(arg, ndarray):
            out.append(torch.from_numpy(arg))
        else:
            out.append(arg)

    return out


class SNRLoss(nn.Module):
    def __init__(self, return_neg=True, eps=1e-8, reduction="mean"):
        super().__init__()
        self.return_neg = return_neg
        self.eps = eps
        self.reduction = reduction

    def forward(self, est: Tensor, ref: Tensor) -> Tensor:
        """Calculate SNR loss.

        Args:
            est: The estimated signal. Shape: [B, T]
            ref: The reference signal. Shape: [B, T]

        Returns:
            The SNR loss value with negative sign. Shape: [B]
        """
        est, ref = numpy_to_torch(est, ref)

        noise = est - ref

        snr = 20 * (
            torch.log10(torch.norm(ref, p=2, dim=1).clamp(min=self.eps))
            - torch.log10(torch.norm(noise, p=2, dim=1).clamp(min=self.eps))
        )

        if self.reduction == "mean":
            snr = snr.mean()
        elif self.reduction == "sum":
            snr = snr.sum()
        elif self.reduction == "none":
            pass
        else:
            raise ValueError(f"Invalid reduction mode: {self.reduction}")

        return -snr if self.return_neg else snr


class SISNRLoss(nn.Module):
    def __init__(self, return_neg=True, eps=1e-8):
        super().__init__()
        self.return_neg = return_neg
        self.eps = eps

    def forward(self, est, ref):
        est, ref = numpy_to_torch(est, ref)

        if est.shape != ref.shape:
            raise RuntimeError(f"Dimension mismatch when calculating SI-SNR, {est.shape=} vs {ref.shape=}")

        s_est = est - torch.mean(est, dim=-1, keepdim=True)
        s_ref = ref - torch.mean(ref, dim=-1, keepdim=True)

        pair_wise_dot = torch.sum(s_ref * s_est, dim=-1, keepdim=True)
        s_ref_norm = torch.sum(s_ref**2, dim=-1, keepdim=True)
        pair_wise_proj = pair_wise_dot * s_ref / (s_ref_norm + self.eps)

        e_noise = s_est - pair_wise_proj

        pair_wise_sdr = torch.sum(pair_wise_proj**2, dim=-1) / (torch.sum(e_noise**2, dim=-1) + self.eps)
        pair_wise_sdr = torch.clamp(pair_wise_sdr, min=self.eps)

        si_snr = torch.mean(10 * torch.log10(pair_wise_sdr))

        return -si_snr if self.return_neg else si_snr


class angle(Function):
    """Similar to torch.angle but robustify the gradient for zero magnitude."""

    @staticmethod
    def forward(ctx, x: Tensor):
        ctx.save_for_backward(x)
        return torch.atan2(x.imag, x.real)

    @staticmethod
    def backward(ctx, grad: Tensor):
        eps = torch.finfo(grad.dtype).eps
        (x,) = ctx.saved_tensors
        grad_inv = grad / (x.real.square() + x.imag.square()).clamp_min_(eps)
        return torch.view_as_complex(torch.stack((-x.imag * grad_inv, x.real * grad_inv), dim=-1))


class MultiResSpecLoss(nn.Module):
    def __init__(self, n_ffts, gamma=1, factor=1, f_complex=None):
        super().__init__()
        self.n_fft_list = n_ffts
        self.gamma = gamma
        self.factor = factor
        self.factor_complex = f_complex

    @staticmethod
    def stft(y, n_fft, hop_length=None, win_length=None):
        num_dims = y.dim()
        assert num_dims == 2, "Only support 2D input."

        hop_length = hop_length or n_fft // 4
        win_length = win_length or n_fft

        batch_size, num_samples = y.size()

        # [B, F, T] in complex-valued
        complex_stft = torch.stft(
            y,
            n_fft,
            hop_length,
            win_length,
            window=torch.hann_window(n_fft, device=y.device),
            return_complex=True,
            normalized=True,
        )

        return complex_stft

    def forward(self, est: Tensor, target: Tensor) -> Tensor:
        eps = torch.finfo(est.dtype).eps
        loss = torch.zeros((), device=est.device, dtype=est.dtype)

        for n_fft in self.n_fft_list:
            Y = self.stft(est, n_fft)
            S = self.stft(target, n_fft)
            Y_abs = Y.abs()
            S_abs = S.abs()
            if self.gamma != 1:
                Y_abs = Y_abs.clamp_min(eps).pow(self.gamma)
                S_abs = S_abs.clamp_min(eps).pow(self.gamma)

            # magnitude loss
            loss += F.mse_loss(Y_abs, S_abs) * self.factor

            # real/imaginary loss
            if self.factor_complex is not None:
                if self.gamma != 1:
                    Y = Y_abs * torch.exp(1j * angle.apply(Y))
                    S = S_abs * torch.exp(1j * angle.apply(S))
                loss += F.mse_loss(torch.view_as_real(Y), torch.view_as_real(S)) * self.factor_complex
        return loss


class CombineLoss(nn.Module):
    def __init__(
        self,
        n_ffts: Iterable[int],
        gamma: float = 1,
        factor: float = 1,
        f_complex: float = None,
    ):
        super().__init__()
        self.n_ffts = n_ffts
        self.gamma = gamma
        self.f = factor
        self.f_complex = f_complex

        self.mulres_loss = MultiResSpecLoss(n_ffts, gamma, factor, f_complex)
        self.l1_loss = nn.L1Loss()

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        loss1 = self.mulres_loss(input, target)
        loss2 = self.l1_loss(input, target)
        return loss1 + loss2


def freq_MAE(estimation, target, win=2048, stride=512, srs=None, sudo_sr=None) -> Tensor:
    est_spec = torch.stft(
        estimation.view(-1, estimation.shape[-1]),
        n_fft=win,
        hop_length=stride,
        window=torch.hann_window(win).to(estimation.device).float(),
        return_complex=True,
    )
    est_target = torch.stft(
        target.view(-1, target.shape[-1]),
        n_fft=win,
        hop_length=stride,
        window=torch.hann_window(win).to(estimation.device).float(),
        return_complex=True,
    )

    if srs is None:
        return (est_spec.real - est_target.real).abs().mean() + (est_spec.imag - est_target.imag).abs().mean()
    else:
        loss = Tensor(0.0)
        for i, sr in enumerate(srs):
            max_freq = int(est_spec.shape[-2] * sr / sudo_sr)
            loss += (est_spec[i][:max_freq].real - est_target[i][:max_freq].real).abs().mean() + (
                est_spec[i][:max_freq].imag - est_target[i][:max_freq].imag
            ).abs().mean()
        loss = loss / len(srs)
        return loss


def mag_MAE(estimation, target, win=2048, stride=512, srs=None, sudo_sr=None) -> Tensor:
    est_spec = torch.stft(
        estimation.view(-1, estimation.shape[-1]),
        n_fft=win,
        hop_length=stride,
        window=torch.hann_window(win).to(estimation.device).float(),
        return_complex=True,
    )
    est_target = torch.stft(
        target.view(-1, target.shape[-1]),
        n_fft=win,
        hop_length=stride,
        window=torch.hann_window(win).to(estimation.device).float(),
        return_complex=True,
    )
    if srs is None:
        return (est_spec.abs() - est_target.abs()).abs().mean()
    else:
        loss = Tensor(0.0)
        for i, sr in enumerate(srs):
            max_freq = int(est_spec.shape[-2] * sr / sudo_sr)
            loss += (est_spec[i][:max_freq].abs() - est_target[i][:max_freq].abs()).abs().mean()
        loss = loss / len(srs)
    return loss


class BinauralLoss(nn.Module):
    def __init__(
        self, n_fft, win_length, hop_length, sr, ild_weight=0.1, ipd_weight=1.0, stoi_weight=0, snr_loss_weight=0.1
    ):
        super().__init__()
        self.stft = partial(stft, n_fft=n_fft, win_length=win_length, hop_length=hop_length)
        self.istft = partial(istft, win_length=win_length, hop_length=hop_length)

        # dynamically import the torch_stoi library
        torch_stoi = importlib.import_module("torch_stoi")
        self.stoi_loss = torch_stoi.NegSTOILoss(sample_rate=sr)
        self.snr_loss = SNRLoss()

        self.ild_weight = ild_weight
        self.ipd_weight = ipd_weight
        self.stoi_weight = stoi_weight
        self.snr_loss_weight = snr_loss_weight

    def _amplitude_to_db(self, x: Tensor, eps: float = 1e-8):
        # Convert amplitude to dB
        return 20 * torch.log10(x.clamp(min=eps))

    def _ild_in_dB(est: Tensor, ref: Tensor, eps=1e-8) -> Tensor:
        """Calculate ILD in dB.

        Args:
            est: the estimated magnitude spectrogram. Shape: [B, F, T]
            ref: the reference magnitude spectrogram. Shape: [B, F, T]
        """
        est_in_db = 20 * torch.log10(est.clamp(min=eps))
        ref_in_db = 20 * torch.log10(ref.clamp(min=eps))

        return est_in_db - ref_in_db

    def _ipd_in_rad(est: Tensor, ref: Tensor) -> Tensor:
        """Calculate IPD in rad.

        Args:
            est: the estimated STFT spectrogram. Shape: [B, F, T]
            ref: the reference STFT spectrogram. Shape: [B, F, T]
        """
        est_phase = torch.angle(est)
        ref_phase = torch.angle(ref)

        return est_phase - ref_phase

    def _build_energy_mask(self, mag_spec_l, mag_spec_r, threshold_in_dB=20):
        """Build speech mask from magnitude spectrogram.

        Assuming a threshold. For each frequency, if the energy of the left and right channel is above the threshold,
        then the mask is 1, otherwise 0. The mask is computed in dB.

        Args:
            mag_spec_l: The magnitude spectrogram of the left channel. Shape: [B, F, T]
            mag_spec_r: The magnitude spectrogram of the right channel. Shape: [B, F, T]
            threshold: The threshold in dB. Default: 15 dB.

        Returns:
            The speech mask (int). Shape: [B, F, T]
        """
        batch_size, num_freqs, num_frames = mag_spec_l.size()

        # Find the the maximum energy computed for each frequency bin
        threshold_l = torch.max(mag_spec_l**2, dim=2)
        threshold_l_in_dB = 10 * torch.log10(threshold_l) - threshold_in_dB
        threshold_l_in_dB = threshold_l_in_dB.unsqueeze(2).repeat(1, 1, num_frames)

        threshold_r = torch.max(mag_spec_r**2, dim=2)
        threshold_r_in_dB = 10 * torch.log10(threshold_r) - threshold_in_dB
        threshold_r_in_dB = threshold_r_in_dB.unsqueeze(2).repeat(1, 1, num_frames)

        # Cho
        binary_mask_l = (self._amplitude_to_db(mag_spec_l) > threshold_l_in_dB).float()
        binary_mask_r = (self._amplitude_to_db(mag_spec_r) > threshold_r_in_dB).float()

        binaural_mask = torch.bitwise_and(binary_mask_l.int(), binary_mask_r.int())

        return binaural_mask

    def forward(self, est, ref):
        # est: [B, 2, T]
        # ref: [B, 2, T]

        # STFT
        est_mag, est_phase, *_ = self.stft(est)  # [B, 2, F, T]
        ref_mag, ref_phase, *_ = self.stft(ref)

        est_mag_l, est_mag_r, est_phase_l, est_phase_r = est_mag[:, 0], est_mag[:, 1], est_phase[:, 0], est_phase[:, 1]
        ref_mag_l, ref_mag_r, ref_phase_l, ref_phase_r = ref_mag[:, 0], ref_mag[:, 1], ref_phase[:, 0], ref_phase[:, 1]

        loss = 0

        if self.snr_loss_weight > 0.0:
            snr_l = self.snr_loss(est[:, 0], ref[:, 0])
            snr_r = self.snr_loss(est[:, 1], ref[:, 1])

            snr_loss = (snr_l + snr_r) / 2

            loss += self.snr_loss_weight * snr_loss

        if self.stoi_weight > 0.0:
            stoi_l = self.stoi_loss(est[:, 0], ref[:, 0])
            stoi_r = self.stoi_loss(est[:, 1], ref[:, 1])

            stoi_loss = (stoi_l + stoi_r) / 2
            stoi_loss = stoi_loss.mean()  # [B] -> []

            loss += self.stoi_weight * stoi_loss

        if self.ild_weight > 0.0:
            est_ild = self._ild_in_dB(est_mag_l, est_mag_r)
            ref_ild = self._ild_in_dB(ref_mag_l, ref_mag_r)

            ild_loss = torch.abs(est_ild - ref_ild)  # [B, F, T]

            binaural_mask = self._build_energy_mask(est_mag_l, est_mag_r)  # [B, F, T]

            masked_ild_loss = (ild_loss * binaural_mask).sum(dim=(1, 2)) / binaural_mask.sum(dim=(1, 2))

            loss += self.ild_weight * masked_ild_loss.mean()

        if self.ipd_weight > 0.0:
            ref_ipd = ref_phase_l - ref_phase_r
            est_ipd = est_phase_l - est_phase_r

            ipd_loss = torch.abs(est_ipd - ref_ipd)

            binaural_mask = self._build_energy_mask(est_mag_l, est_mag_r, threshold_in_dB=20)  # [B, F, T]

            masked_ipd_loss = (ipd_loss * binaural_mask).sum(dim=(1, 2)) / binaural_mask.sum(dim=(1, 2))

            loss += self.ipd_weight * masked_ipd_loss.mean()

        return loss


if __name__ == "__main__":
    n_ffts = [240, 480, 960, 1440]
    gamma = 0.3
    factor = 1
    f_complex = 1

    loss = CombineLoss(n_ffts, gamma, factor, f_complex)
    input = torch.rand(2, 16000)
    target = torch.rand(2, 16000)
    print(loss(input, target))



from typing import Union

import numpy as np
import torch

from audiozen.metrics.metric_utils import Metric


class SISDR:
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        estimate: Union[torch.tensor, np.ndarray],
        target: Union[torch.tensor, np.ndarray],
        reduce_mean: bool = True,
    ) -> dict:
        if isinstance(estimate, np.ndarray):
            estimate = torch.tensor(estimate)

        if isinstance(target, np.ndarray):
            target = torch.tensor(target)

        eps = torch.finfo(estimate.dtype).eps

        # zero mean to ensure scale invariance
        s_target = target - torch.mean(target, dim=-1, keepdim=True)
        s_estimate = estimate - torch.mean(estimate, dim=-1, keepdim=True)

        # <s, s'> / ||s||**2 * s
        pair_wise_dot = torch.sum(s_target * s_estimate, dim=-1, keepdim=True)
        s_target_norm = torch.sum(s_target**2, dim=-1, keepdim=True)
        pair_wise_proj = (pair_wise_dot * s_target + eps) / (s_target_norm + eps)

        e_noise = s_estimate - pair_wise_proj

        pair_wise_sdr = (torch.sum(pair_wise_proj**2, dim=-1) + eps) / (torch.sum(e_noise**2, dim=-1) + eps)
        val = 10 * torch.log10(pair_wise_sdr + eps)

        if reduce_mean:
            val = torch.mean(val)

        val = val.detach().item()

        return Metric(name="si_sdr", value=val)



import logging
import os
import platform
import shutil
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
import psutil
import torch
from deprecated import deprecated


logger = logging.getLogger(__name__)


@dataclass
class StateName:
    ADAPTER: str = "adapter"
    MODEL: str = "model"
    OPTIMIZER: str = "optimizer"
    LR_SCHEDULER: str = "lr_scheduler"
    SEED: str = "seed"
    TRAINER_STATE: str = "trainer_state"
    SCALER: str = "scaler"


def has_length(dataset):
    """Checks if the dataset implements __len__() and it doesn't raise an error."""
    try:
        return len(dataset) is not None
    except TypeError:
        # TypeError: len() of unsized object
        return False


class TrainerState:
    """State of the Trainer.

    This can be registered by Accelerate for `save_state` and `load_state`.
    """

    def __init__(self, greater_is_better) -> None:
        self.epochs_trained = 0
        self.steps_trained = 0

        self.early_stopping_patience_counter = 0

        self.best_score = -np.inf if greater_is_better else np.inf
        self.best_score_epoch = 0

    def load_state_dict(self, state_dict: dict) -> None:
        self.epochs_trained = state_dict["epochs_trained"]
        self.steps_trained = state_dict["steps_trained"]

        self.early_stopping_patience_counter = state_dict["early_stopping_patience_counter"]

        self.best_score = state_dict["best_score"]
        self.best_score_epoch = state_dict["best_score_epoch"]

    def state_dict(self) -> dict:
        return {
            "epochs_trained": self.epochs_trained,
            "steps_trained": self.steps_trained,
            "early_stopping_patience_counter": self.early_stopping_patience_counter,
            "best_score": self.best_score,
            "best_score_epoch": self.best_score_epoch,
        }


def collect_env():
    pt_version = torch.__version__
    pt_cuda_available = torch.cuda.is_available()

    info = {
        "Platform": platform.platform(),
        "Python version": platform.python_version(),
        "Numpy version": np.__version__,
        "PyTorch version (GPU?)": f"{pt_version} ({pt_cuda_available})",
        "System RAM": f"{psutil.virtual_memory().total / 1024**3:.2f} GB",
        "GPU Available": f"{pt_cuda_available}",
        "GPU IDs": f"{torch.cuda.device_count()}",
    }

    if pt_cuda_available:
        info["GPU type"] = torch.cuda.get_device_name()

    return "\n".join([f"- {prop}: {val}" for prop, val in info.items()])


def print_commandline_args():
    return " ".join(sys.argv)


def setup_ckpt_dir_for_epoch_i(ckpt_root_dir: Path, epoch: int):
    ckpt_epoch_dir = ckpt_root_dir / f"epoch_{str(epoch).zfill(4)}"
    ckpt_epoch_dir.mkdir(parents=True, exist_ok=True)
    return ckpt_epoch_dir


@deprecated(
    version="2025.1.21",
    reason="We follow the API design of Huggingface Transfroms. They directly save the model state dict in the `trainer.py`.",
)
def save_state_dict_to_checkpoint(ckpt_epoch_dir, state_dict: dict[str, Any]):
    """Save the state dictionary to the checkpoint directory.

    Note:
        For the `model` key in the state list, it can be a single model or a list of models (e.g., GAN).
        Please call this function on the rank 0 process.

    Args:
        ckpt_epoch_dir: The directory to save the checkpoint of the current epoch.
        state_dict: The state dictionary to save.
    """
    ckpt_epoch_dir.mkdir(parents=True, exist_ok=True)

    # Model states
    if StateName.MODEL in state_dict and state_dict[StateName.MODEL] is not None:
        model_states = state_dict[StateName.MODEL]

        if isinstance(model_states, list):
            for i, model_state in enumerate(model_states):
                state_fpath = ckpt_epoch_dir / f"model_{i}.pth"
                torch.save(model_state, state_fpath)
                logger.info(f"Model checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")
        else:
            state_fpath = ckpt_epoch_dir / "model.pth"
            torch.save(state_dict[StateName.MODEL], state_fpath)
            logger.info(f"Model checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")

    # Adapter states
    if StateName.ADAPTER in state_dict and state_dict[StateName.ADAPTER] is not None:
        state_fpath = ckpt_epoch_dir / "adapter.pth"
        state_weights = state_dict[StateName.ADAPTER]
        torch.save(state_weights, state_fpath)
        logger.info(f"Adapter checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")

    # Optimizer states
    if StateName.OPTIMIZER in state_dict and state_dict[StateName.OPTIMIZER] is not None:
        state_fpath = ckpt_epoch_dir / "optimizer.pth"
        state_weights = state_dict[StateName.OPTIMIZER]
        torch.save(state_weights, state_fpath)
        logger.info(f"Optimizer checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")

    # LR Scheduler states
    if StateName.LR_SCHEDULER in state_dict and state_dict[StateName.LR_SCHEDULER] is not None:
        state_fpath = ckpt_epoch_dir / "lr_scheduler.pth"
        torch.save(state_dict[StateName.LR_SCHEDULER], state_fpath)
        logger.info(f"LR Scheduler checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")

    # Trainer states
    if StateName.TRAINER_STATE in state_dict and state_dict[StateName.TRAINER_STATE] is not None:
        state_fpath = ckpt_epoch_dir / "trainer_state.pkl"
        torch.save(state_dict[StateName.TRAINER_STATE], state_fpath)
        logger.info(f"Trainer state checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")

    if StateName.SCALER in state_dict and state_dict[StateName.SCALER] is not None:
        state_fpath = ckpt_epoch_dir / "scaler.pkl"
        torch.save(state_dict[StateName.SCALER], state_fpath)
        logger.info(f"Scaler checkpoint {os.path.getsize(state_fpath) / 1000**3:.2f} GB saved to {state_fpath}")


@deprecated(
    version="2025.1.21",
    reason="We follow the API design of Huggingface Transfroms. They directly save the model state dict in the `trainer.py`.",
)
def load_state_dict_from_checkpoint(ckpt_epoch_dir: Path) -> dict[str, Any]:
    """Load all possible states from the checkpoint directory (e.g., model, adapter, optimizer, lr_scheduler, trainer).

    Note:
        For the `model` key in the state list, it can be a single model or a dictionary of models (e.g., GAN).
        Please call this function on the rank 0 process.

    Args:
        ckpt_epoch_dir: The directory to load the checkpoint of the current epoch.

    Returns:
        The state dictionary loaded from the checkpoint directory.
    """
    state_dict = {}

    # Model states
    model_state_fpaths = list(ckpt_epoch_dir.glob("model*.pth"))

    if len(model_state_fpaths) > 1:
        model_states = {}
        for model_state_fpath in model_state_fpaths:
            model_name = model_state_fpath.stem
            model_states[model_name] = torch.load(model_state_fpath, map_location="cpu", weights_only=True)
        state_dict[StateName.MODEL] = model_states
    elif len(model_state_fpaths) == 1:
        state_dict[StateName.MODEL] = torch.load(model_state_fpaths[0], map_location="cpu", weights_only=True)

    # Adapter states
    adapter_state_fpath = ckpt_epoch_dir / "adapter.pth"
    if adapter_state_fpath.exists():
        state_dict[StateName.ADAPTER] = torch.load(adapter_state_fpath, map_location="cpu", weights_only=True)

    # Optimizer states
    optimizer_state_fpath = ckpt_epoch_dir / "optimizer.pth"
    if optimizer_state_fpath.exists():
        state_dict[StateName.OPTIMIZER] = torch.load(optimizer_state_fpath, map_location="cpu", weights_only=True)

    # LR Scheduler states
    lr_scheduler_state_fpath = ckpt_epoch_dir / "lr_scheduler.pth"
    if lr_scheduler_state_fpath.exists():
        state_dict[StateName.LR_SCHEDULER] = torch.load(
            lr_scheduler_state_fpath, map_location="cpu", weights_only=True
        )

    # Trainer states
    trainer_state_fpath = ckpt_epoch_dir / "trainer_state.pkl"
    if trainer_state_fpath.exists():
        state_dict[StateName.TRAINER_STATE] = torch.load(trainer_state_fpath, weights_only=False)

    # Scaler states
    scaler_state_fpath = ckpt_epoch_dir / "scaler.pkl"
    if scaler_state_fpath.exists():
        state_dict[StateName.SCALER] = torch.load(scaler_state_fpath, weights_only=True)

    return state_dict


def cleanup_redundant_checkpoints(ckpt_root_dir: Path, num_keep_ckpts: int):
    ckpt_epoch_dirs = sorted(ckpt_root_dir.glob("epoch_*"), key=lambda x: int(x.stem.split("_")[-1]), reverse=True)

    for ckpt_epoch_dir in ckpt_epoch_dirs[num_keep_ckpts:]:
        shutil.rmtree(ckpt_epoch_dir)
        logger.info(f"Checkpoint {ckpt_epoch_dir} is removed.")

    logger.info(f"Kept the latest {num_keep_ckpts} checkpoints.")



from dataclasses import dataclass, field
from functools import partial
from typing import List, Optional

import torch
import torch.nn as nn
from einops import rearrange
from simple_parsing import Serializable
from torch.nn import functional as F

from audiozen.acoustics.audio_feature import istft, stft
from audiozen.models.spiking_fullsubnet.efficient_spiking_neuron import MemoryState, efficient_spiking_neuron


class SequenceModel(nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        sequence_model="GSN",
        proj_size=0,
        shared_weights=False,
        output_activate_function=None,
        bn=False,
        use_pre_layer_norm=True,
    ):
        super().__init__()

        if use_pre_layer_norm:
            self.pre_layer_norm = nn.LayerNorm(input_size)

        if sequence_model == "GSN":
            self.sequence_model = efficient_spiking_neuron(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                shared_weights=shared_weights,
                bn=bn,
            )
        else:
            raise NotImplementedError(f"Sequence model {sequence_model} not implemented.")

        if proj_size > 0:
            self.proj = nn.Linear(hidden_size, proj_size)
        else:
            self.proj = nn.Identity()

        if output_activate_function == "tanh":
            self.output_activate_function = nn.Tanh()
        elif output_activate_function == "sigmoid":
            self.output_activate_function = nn.Sigmoid()
        elif output_activate_function == "relu":
            self.output_activate_function = nn.ReLU()
        else:
            self.output_activate_function = nn.Identity()

        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.use_pre_layer_norm = use_pre_layer_norm

    def forward(self, input):
        """Forward function.

        Args:
            input (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)` or `(batch_size, num_channels, num_freq, sequence_length)`):
                Input 3D or 4D tensor.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)`): Output tensor.
                Output tensor.
        """
        assert input.ndim == 3, f"Input tensor must be 3D, but got {input.ndim}D."

        batch_size, num_freqs, sequence_length = input.shape

        # Initialize memory states.
        states = [
            MemoryState(
                torch.zeros(batch_size, self.hidden_size, device=input.device),
                torch.zeros(batch_size, self.hidden_size, device=input.device),
            )
            for _ in range(self.num_layers)
        ]

        input = rearrange(input, "b f t -> t b f")

        # Apply layer normalization.
        if self.use_pre_layer_norm:
            input = self.pre_layer_norm(input)

        # Pass through the sequence model.
        output, _, all_layer_outputs = self.sequence_model(input, states)

        # Project the output if necessary.
        output = self.proj(output)
        all_layer_outputs += [output]

        # Apply the output activation function.
        output = self.output_activate_function(output)

        output = rearrange(output, "t b f -> b f t")
        return output, all_layer_outputs


class SubBandSequenceModel(SequenceModel):
    def __init__(self, df_order, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.df_order = df_order

    def forward(self, input_features):
        """How to process the subband features.

        - `fs`: number of frequency bins in the subband.
        - `fc`: number of frequency bins in the center frequency.
        - `df`: order of the deep filter.
        - `n`: number of subbands.

        Args:
            input (`torch.Tensor` of shape `(batch_size, num_subbands, num_channels, sb_freq_size, sequence_length)`):
                Subband feature at a particular frequency.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, df_order, num_subbands * ctr_freq_size, sequence_length, num_channels)`):
                Complex output tensor.
        """
        batch_size, num_subbands, num_channels, sb_freq_size, sequence_length = input_features.shape
        assert num_channels == 1, "Only mono audio is supported."

        input_features = rearrange(input_features, "b n c fs t -> (b n) (c fs) t")

        output, all_layer_outputs = super().forward(input_features)

        # `2 * num_channels` because we have real and imaginary parts in the output.
        output = rearrange(
            output, "(b n) (c fc df) t -> b df (n fc) t c", b=batch_size, c=num_channels * 2, df=self.df_order
        )

        return output, all_layer_outputs


class SubbandModel(nn.Module):
    def __init__(
        self,
        freq_cutoffs,
        center_freq_sizes,
        neighbor_freq_sizes,
        df_orders,
        **kwargs,
    ):
        """Subband model.

        Args:
            freq_cutoffs (`list` of `int`):
                 Cutoff frequencies for the subbands. The first and last elements are the lower and upper cutoffs.
            center_freq_sizes (`list` of `int`):
                Number of frequency bins in the center frequency for each subband. The length of this list must be
                equal to the length of `freq_cutoffs` minus 1.
            neighbor_freq_sizes: like `sb_center_freq_sizes`, but for the neighboring frequency bins.
            df_orders: like `center_freq_sizes`, but for the deep filter order.
            kwargs: other arguments for `SequenceModel`.
        """
        super().__init__()
        assert len(freq_cutoffs) - 1 == len(center_freq_sizes), "Number of subbands must be equal to len(cutoffs)."

        sb_models = []
        for ctr_freq, nbr_freq, df_order in zip(center_freq_sizes, neighbor_freq_sizes, df_orders):
            sb_models.append(
                SubBandSequenceModel(
                    input_size=(ctr_freq + nbr_freq * 2) + ctr_freq,
                    proj_size=ctr_freq * df_order * 2,
                    df_order=df_order,
                    **kwargs,
                )
            )

        self.sb_models = nn.ModuleList(sb_models)
        self.freq_cutoffs = freq_cutoffs
        self.center_freq_sizes = center_freq_sizes
        self.neighbor_freq_sizes = neighbor_freq_sizes
        self.df_orders = df_orders

    def forward(self, noisy_input, fb_output):
        """Frequency-wise processing of the subband features.

        Separate the noisy input into several sections. Each section has the same center frequency and neighboring
        frequency bins. Then, process each section with the corresponding subband model.

        For a subband feature at frequency `f`, the corresponding feature is obtained by concatenating:
        1. `f` itself.
        2. `f - N` to `f - 1`, where `N` is the number of frequency bins on each side of `f`.
        3. `f + 1` to `f + N`.
        4. corresponding frequency bins in the fullband feature.

        Args:
            noisy_input (`torch.Tensor` of shape `(batch_size, num_channels, num_freqs, num_frames)`):
                Noisy input spectrogram.
            fb_output (`torch.Tensor` of shape `(batch_size, num_channels, num_freqs, num_frames)`):
                Fullband output spectrogram.
        """
        batch_size, num_channels, num_freqs, num_frames = noisy_input.size()
        assert num_channels == 1, "Only mono audio is supported."

        output = []
        all_layer_outputs = []
        for idx, sb_model in enumerate(self.sb_models):
            # [batch_size, num_subbands, num_channels, sb_freq_size, num_frames]
            noisy_subbands = self._freq_unfold(
                input=noisy_input,
                lower_cutoff_freq=self.freq_cutoffs[idx],
                upper_cutoff_freq=self.freq_cutoffs[idx + 1],
                ctr_freq=self.center_freq_sizes[idx],
                nbr_freq=self.neighbor_freq_sizes[idx],
            )

            fb_subbands = self._freq_unfold(
                input=fb_output,
                lower_cutoff_freq=self.freq_cutoffs[idx],
                upper_cutoff_freq=self.freq_cutoffs[idx + 1],
                ctr_freq=self.center_freq_sizes[idx],
                nbr_freq=0,
            )

            # Concatenate the subband features with the corresponding fullband features.
            sb_input = torch.cat([noisy_subbands, fb_subbands], dim=-2)
            sb_output, sb_all_layer_outputs = sb_model(sb_input)
            output += [sb_output]
            all_layer_outputs += [sb_all_layer_outputs]

        return output, all_layer_outputs

    def _freq_unfold(self, input, lower_cutoff_freq, upper_cutoff_freq, ctr_freq, nbr_freq):
        """Unfold the frequency bins based on a given lower and upper cutoff frequency bondaries.

        Args:
            input (`torch.Tensor` of shape `(batch_size, num_channels, num_freqs, num_frames)`):
                Noisy input spectrogram.
            lower_cutoff_freq: lower cutoff frequency of current section.
            upper_cutoff_freq: upper cutoff frequency of current section.
            ctr_freq: number of frequency bins in the center frequency.
            nbr_freq: number of neighboring frequency bins.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, num_subbands, num_channels, sb_freq_size, num_frames)`):
                Unfolded tensor.
        """
        batch_size, num_channels, num_freqs, num_frames = input.size()
        assert num_channels == 1, "Only mono audio is supported."

        if (upper_cutoff_freq - lower_cutoff_freq) % ctr_freq != 0:
            raise ValueError(
                f"Number of frequency bins must be divisible by the center frequency."
                f"GOT: {ctr_freq=}, {upper_cutoff_freq=}, {lower_cutoff_freq=}"
            )

        # extract valid input with the shape of [batch_size, 1, num_subbands * ctr_freq + 2 * nbr_freq, num_frames]
        if lower_cutoff_freq == 0:
            # lower_cutoff_freq = 0 is a special case.
            # lower = 0, upper = upper_cutoff_freq + nbr_freq
            valid_input = input[..., : upper_cutoff_freq + nbr_freq, :]
            valid_input = F.pad(valid_input, (0, 0, nbr_freq, 0), mode="reflect")
        elif upper_cutoff_freq == num_freqs:
            # upper_cutoff_freq = num_freqs is a special case.
            # lower = lower_cutoff_freq - nbr_freq, upper = num_freqs
            valid_input = input[..., lower_cutoff_freq - nbr_freq :, :]
            valid_input = F.pad(valid_input, (0, 0, 0, nbr_freq), mode="reflect")
        else:
            # lower = lower_cutoff_freq - nbr_freq, upper = upper_cutoff_freq + nbr_freq
            valid_input = input[..., lower_cutoff_freq - nbr_freq : upper_cutoff_freq + nbr_freq, :]

        # Unfold the frequency bins.
        output = F.unfold(
            input=valid_input, kernel_size=(ctr_freq + nbr_freq * 2, num_frames), stride=(ctr_freq, num_frames)
        )

        # Reshape the output to [batch_size, num_subbands, num_channels, sb_freq_size, num_frames].
        output = rearrange(output, "b (c fs t) n -> b n c fs t", c=num_channels, fs=ctr_freq + nbr_freq * 2)

        return output


def deepfiltering(complex_spec, coef, order: int):
    """Deep filtering implementation using `torch.einsum`. Requires unfolded spectrogram.

    Args:
        complex_spec (`torch.ComplexTensor` of shape `[B, C, F, T]`):
            Complex spectrogram.
        coef (`torch.Tensor` of shape `[B, C * order, F, T, 2]`):
            Coefficients of the deep filter.
        order (`int`): Order of the deep filter.

    Returns:
        spec (complex Tensor): Spectrogram of shape [B, C, F, T]
    """
    need_unfold = order > 1

    if need_unfold:
        complex_spec = F.pad(complex_spec, (order - 1, 0))
        complex_spec = complex_spec.unfold(3, order, 1)  # [B, C, F, T, df]
    else:
        complex_spec = complex_spec.unsqueeze(-1)  # [B, C, F, T, 1]

    complex_coef = torch.complex(coef[..., 0], coef[..., 1])  # [B, C * df, F, T]
    complex_coef = rearrange(complex_coef, "b (c df) f t -> b c df f t", df=order)

    # df
    out = torch.einsum("...ftn,...nft->...ft", complex_spec, complex_coef)

    return out


@dataclass
class ModelArgs(Serializable):
    """Options for the model."""

    n_fft: int = 512
    hop_length: int = 128
    win_length: int = 512
    fdrc: float = 0.5
    fb_input_size: int = 64
    fb_hidden_size: int = 320
    fb_num_layers: int = 2
    fb_proj_size: int = 64
    fb_output_activate_function: Optional[str] = None
    sb_hidden_size: int = 224
    sb_num_layers: int = 2
    freq_cutoffs: List[int] = field(default_factory=lambda: [0, 32, 128, 256])
    df_orders: List[int] = field(default_factory=lambda: [5, 3, 1])
    center_freq_sizes: List[int] = field(default_factory=lambda: [2, 32, 64])
    neighbor_freq_sizes: List[int] = field(default_factory=lambda: [15, 15, 15])
    use_pre_layer_norm_fb: bool = True
    use_pre_layer_norm_sb: bool = True
    bn: bool = False
    shared_weights: bool = False
    sequence_model: str = "GSN"
    num_spks: int = 1


class SpikingFullSubNet(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.args = args

        self.fb_model = SequenceModel(
            input_size=args.fb_input_size,
            hidden_size=args.fb_hidden_size,
            num_layers=args.fb_num_layers,
            shared_weights=args.shared_weights,
            sequence_model=args.sequence_model,
            proj_size=args.fb_proj_size,
            output_activate_function=args.fb_output_activate_function,
            bn=args.bn,
            use_pre_layer_norm=args.use_pre_layer_norm_fb,
        )

        self.sb_model = SubbandModel(
            freq_cutoffs=args.freq_cutoffs,
            center_freq_sizes=args.center_freq_sizes,
            neighbor_freq_sizes=args.neighbor_freq_sizes,
            df_orders=args.df_orders,
            hidden_size=args.sb_hidden_size,
            num_layers=args.sb_num_layers,
            shared_weights=args.shared_weights,
            sequence_model=args.sequence_model,
            bn=args.bn,
            use_pre_layer_norm=args.use_pre_layer_norm_sb,
        )

        self.subband_model = None

        self.stft = partial(stft, n_fft=args.n_fft, hop_length=args.hop_length, win_length=args.win_length)
        self.istft = partial(istft, n_fft=args.n_fft, hop_length=args.hop_length, win_length=args.win_length)

        self.fb_input_size = args.fb_input_size
        self.n_fft = args.n_fft
        self.hop_length = args.hop_length
        self.win_length = args.win_length
        self.fdrc = args.fdrc
        self.df_orders = args.df_orders

    def forward(self, input):
        """Forward function.

        Args:
            input (`torch.Tensor` of shape `(batch_size, sequence_length)`):
                Waveform tensor.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)`):
                Output tensor.
        """
        assert input.ndim == 2, f"Input tensor must be 2D, but got {input.ndim}D."
        batch_size, sequence_length = input.shape

        noisy_mag, _, noisy_real, noisy_imag = self.stft(input)
        noisy_cmp = torch.complex(real=noisy_real, imag=noisy_imag)
        noisy_cmp = rearrange(noisy_cmp, "b f t -> b 1 f t")

        # ================== Fullband model ==================
        noisy_mag = rearrange(noisy_mag, "b f t -> b 1 f t")
        noisy_mag = noisy_mag**self.fdrc
        noisy_mag = noisy_mag[..., :-1, :]

        # Extract the lowband signal.
        fb_input = noisy_mag[..., : self.fb_input_size, :]
        fb_input = rearrange(fb_input, "b c f t -> b (c f) t")
        fb_output, fb_all_layer_outputs = self.fb_model(fb_input)
        fb_output = rearrange(fb_output, "b f t -> b 1 f t")
        fb_output = fb_output.repeat(1, 1, (self.n_fft // 2 + 1) // self.fb_input_size, 1)

        # ================== Subband model ==================
        # list [[B, df, F_1, T, 2], [B, df, F_2, T, 2], ...], where F = F_1 + F_2 + ...
        df_coef_list, sb_all_layer_outputs = self.sb_model(noisy_mag, fb_output)

        # ================== Reconstruct the output ==================
        num_filtered_freqs = 0
        enh_freqs_list = []
        for df_coef, df_order in zip(df_coef_list, self.df_orders):
            num_freqs = df_coef.shape[2]
            comp_stft_in = noisy_cmp[..., num_filtered_freqs : num_filtered_freqs + num_freqs, :]
            enh_freqs = deepfiltering(comp_stft_in, df_coef, df_order)
            enh_freqs_list.append(enh_freqs)
            num_filtered_freqs += num_freqs

        enh_freqs = torch.cat(enh_freqs_list, dim=2)
        enh_stft = noisy_cmp.clone()
        enh_stft[..., :-1, :] = enh_freqs
        enh_stft = rearrange(enh_stft, "b 1 f t -> b f t")
        enh_mag = torch.abs(enh_stft)  # For computing DNSMOS loss

        enh_y = self.istft(enh_stft, length=sequence_length)

        return enh_y, enh_mag, fb_all_layer_outputs, sb_all_layer_outputs


if __name__ == "__main__":
    from torchinfo import summary

    model = SpikingFullSubNet(
        args=ModelArgs(
            n_fft=512,
            hop_length=128,
            win_length=512,
            fdrc=0.5,
            fb_input_size=64,
            fb_hidden_size=320,
            fb_num_layers=2,
            fb_proj_size=64,
            fb_output_activate_function=None,
            sb_hidden_size=224,
            sb_num_layers=2,
            freq_cutoffs=[0, 32, 128, 256],
            df_orders=[3, 1, 1],
            center_freq_sizes=[8, 32, 64],
            neighbor_freq_sizes=[15, 15, 15],
            use_pre_layer_norm_fb=True,
            use_pre_layer_norm_sb=True,
            bn=True,
            shared_weights=True,
            sequence_model="GSN",
            num_spks=1,
        )
    )

    input = torch.rand(5, 16000)
    # enh_y, enh_mag, fb_all_layer_outputs, sb_all_layer_outputs = model(input)
    # print(f"{len(fb_all_layer_outputs)=}")

    summary(model, input_size=(5, 16000))



from pathlib import Path

import librosa
import numpy as np
import soundfile as sf
from joblib import Parallel, delayed
from torch.utils import data
from tqdm import tqdm

from audiozen.acoustics.audio_feature import load_wav


class BaseDataset(data.Dataset):
    def __init__(self):
        super().__init__()

    @staticmethod
    def _load_dataset_from_text_and_dir_list(dataset_list):
        """Load dataset from text file or directory.

        Args:
            dataset_list: A list of dataset path, which can be a directory or a text file. `dataset_list` can also be a single path of directory or text file.

        Raises:
            FileNotFoundError: If the dataset path is not found.

        Returns:
            A list of file path.
        """
        if isinstance(dataset_list, str):
            dataset_list = [dataset_list]

        fpath_list = []

        for dataset in dataset_list:
            dataset = Path(dataset).expanduser().absolute()
            if dataset.is_dir():
                fpath_list += librosa.util.find_files(dataset.as_posix(), ext="wav")
            elif dataset.is_file():
                with open(dataset, "r") as f:
                    fpath_list += [line.rstrip("\n") for line in f]
            else:
                raise FileNotFoundError(f"File {dataset} not found.")

        return fpath_list

    @staticmethod
    def _load_dataset_in_txt(dataset_path, return_empty_if_not_exist=False):
        dataset_path = Path(dataset_path).expanduser().absolute()
        if dataset_path.is_file():
            with open(dataset_path, "r") as f:
                data = [line.rstrip("\n") for line in f]
                return data
        else:
            if return_empty_if_not_exist:
                return []
            else:
                raise FileNotFoundError(f"File {dataset_path} not found.")

    @staticmethod
    def _offset_and_limit(dataset_list, offset, limit):
        dataset_list = dataset_list[offset:]
        if limit:
            dataset_list = dataset_list[:limit]
        return dataset_list

    @staticmethod
    def _parse_snr_range(snr_range):
        assert len(snr_range) == 2, f"The range of SNR should be [low, high], not {snr_range}."
        assert snr_range[0] <= snr_range[-1], "The low SNR should not larger than high SNR."

        low, high = snr_range
        snr_list = []
        for i in range(low, high + 1, 1):
            snr_list.append(i)

        return snr_list

    def _preload_dataset(self, file_path_list, remark=""):
        waveform_list = Parallel(n_jobs=self.num_workers)(
            delayed(self._load_wav)(f_path, sr=self.sr) for f_path in tqdm(file_path_list, desc=remark)
        )
        return list(zip(file_path_list, waveform_list))  # type: ignore

    def _select_noise_y(self, target_length):
        noise_y = np.zeros(0, dtype=np.float32)
        silence = np.zeros(int(self.sr * self.silence_length), dtype=np.float32)
        remaining_length = target_length

        while remaining_length > 0:
            noise_file = self._random_select_from(self.noise_path_list)
            noise_new_added = load_wav(noise_file, sr=self.sr)
            noise_y = np.append(noise_y, noise_new_added)
            remaining_length -= len(noise_new_added)

            # If still need to add new noise, insert a small silence segment firstly
            if remaining_length > 0:
                silence_len = min(remaining_length, len(silence))
                noise_y = np.append(noise_y, silence[:silence_len])
                remaining_length -= silence_len

        if len(noise_y) > target_length:
            idx_start = np.random.randint(len(noise_y) - target_length)
            noise_y = noise_y[idx_start : idx_start + target_length]

        return noise_y

    @staticmethod
    def _load_wav(path, duration=None, sr=None, mode="wrap"):
        if isinstance(path, Path):
            path = path.as_posix()

        with sf.SoundFile(path) as sf_desc:
            orig_sr = sf_desc.samplerate

            if duration is not None:
                frame_orig_duration = sf_desc.frames
                frame_duration = int(duration * orig_sr)
                if frame_duration < frame_orig_duration:
                    # Randomly select a segment
                    offset = np.random.randint(frame_orig_duration - frame_duration)
                    sf_desc.seek(offset)
                    y = sf_desc.read(frames=frame_duration, dtype=np.float32, always_2d=True).T
                else:
                    y = sf_desc.read(dtype=np.float32, always_2d=True).T  # [C, T]
                    y = np.pad(y, ((0, 0), (0, frame_duration - frame_orig_duration)), mode)
            else:
                y = sf_desc.read(dtype=np.float32, always_2d=True).T

        if y.shape[0] == 1:
            y = y[0]

        if sr is not None:
            y = librosa.resample(y, orig_sr=orig_sr, target_sr=sr)
        else:
            sr = orig_sr

        return y, sr



from typing import List, Tuple

import torch
import torch.nn as nn
from torch import Tensor, jit


class MogrifierLSTMCell(jit.ScriptModule):
    def __init__(self, input_size, hidden_size, mogrify_steps):
        super().__init__()
        self.mogrify_steps = mogrify_steps

        self.lstm = nn.LSTMCell(input_size, hidden_size)

        self.mogrifier_list = nn.ModuleList(
            [
                nn.Linear(hidden_size, input_size),  # q
                nn.Linear(input_size, hidden_size),  # r
                nn.Linear(hidden_size, input_size),  # q
                nn.Linear(input_size, hidden_size),  # r
                nn.Linear(hidden_size, input_size),  # q
            ]
        )

    @jit.script_method
    def mogrify(self, x, h):
        # type: (Tensor, Tensor) -> Tuple[Tensor, Tensor]
        x = (2 * torch.sigmoid(self.mogrifier_list[0](h))) * x
        h = (2 * torch.sigmoid(self.mogrifier_list[1](x))) * h
        x = (2 * torch.sigmoid(self.mogrifier_list[2](h))) * x
        h = (2 * torch.sigmoid(self.mogrifier_list[3](x))) * h
        x = (2 * torch.sigmoid(self.mogrifier_list[4](h))) * x

        return x, h

    @jit.script_method
    def forward(self, x, states):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        ht, ct = states
        x, ht = self.mogrify(x, ht)
        ht, ct = self.lstm(x, (ht, ct))
        return ht, (ht, ct)


class MogrifierLSTMLayer(jit.ScriptModule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.cell = MogrifierLSTMCell(*args, **kwargs)

    @jit.script_method
    def forward(self, input, state):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        input = input.transpose(0, 1)  # [T, B, F]

        outputs = []
        for i in range(len(input)):
            out, state = self.cell(input[i], state)
            outputs += [out]

        outputs = torch.stack(outputs, dim=0)  # [T, B, F]
        outputs = outputs.transpose(0, 1).contiguous()  # [B, T, F]
        return outputs, state


class MogrifierLSTM(jit.ScriptModule):
    def __init__(self, input_size, hidden_size, num_layers, mogrify_steps=4) -> None:
        super().__init__()

        layers = [MogrifierLSTMLayer(input_size, hidden_size, mogrify_steps)]
        for _ in range(num_layers - 1):
            layers.append(MogrifierLSTMLayer(hidden_size, hidden_size, mogrify_steps))

        self.layers = nn.ModuleList(layers)

        self.num_layers = num_layers
        self.hidden_size = hidden_size

    @jit.script_method
    def forward(self, input, states):
        # type: (Tensor, List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]
        # List[LSTMState]: One state per layer
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])

        output = input
        h_0, c_0 = states

        for i, layer in enumerate(self.layers):
            in_state = (h_0[i], c_0[i])  # [B, H]
            output, out_state = layer(output, in_state)  # [B, T, H]
            output_states += [out_state]

        return output, output_states


class MogrifierLSTMWrapper(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers, mogrify_steps=4):
        super().__init__()

        self.mogrifier_lstm = MogrifierLSTM(input_size, hidden_size, num_layers, mogrify_steps=mogrify_steps)
        self.num_layers = num_layers
        self.hidden_size = hidden_size

    def forward(self, input, states=None):
        batch_size = input.shape[0]

        if states is None:
            h_0 = torch.zeros(
                self.num_layers,
                batch_size,
                self.hidden_size,
                device=input.device,
                dtype=input.dtype,
            )
            c_0 = torch.zeros(
                self.num_layers,
                batch_size,
                self.hidden_size,
                device=input.device,
                dtype=input.dtype,
            )
            states = (h_0, c_0)

        output, output_states = self.mogrifier_lstm(input, states)

        return output, output_states


if __name__ == "__main__":
    model = MogrifierLSTMWrapper(
        input_size=257,
        hidden_size=512,
        num_layers=2,
        mogrify_steps=4,
    )

    input = torch.rand(1, 100, 257)
    output, states = model(input)
    print(output.shape)



import numpy as np
from numpy.typing import NDArray


def hz_to_bark(hz: float | NDArray) -> float | NDArray:
    return 26.81 / (1 + 1960.0 / hz) - 0.53


def bark_to_hz(bark: float | NDArray) -> float | NDArray:
    return 1960.0 / (26.81 / (0.53 + bark) - 1)


def bark_filter_bank(
    num_filters: int,
    n_fft: int,
    sr: int,
    low_freq: float,
    high_freq: float,
) -> NDArray:
    high_freq = high_freq or sr / 2
    assert high_freq <= sr / 2, "highfreq is greater than samplerate/2"

    low_bark = hz_to_bark(low_freq)
    high_bark = hz_to_bark(high_freq)
    barkpoints = np.linspace(low_bark, high_bark, num_filters + 2)
    bin = np.floor((n_fft + 1) * bark_to_hz(barkpoints) / sr)
    # bin = np.array(
    #     [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 34, 36, 38, 40, 42,
    #      44, 46, 48, 56, 64, 72, 80, 92, 104, 116, 128, 144, 160, 176, 192, 208, 232, 256])

    print(bin.shape)

    fbank = np.zeros([num_filters, n_fft // 2 + 1])
    for j in range(0, num_filters):
        print(j)
        for i in range(int(bin[j]), int(bin[j + 1])):
            fbank[j, i] = (i - bin[j]) / (bin[j + 1] - bin[j])
        for i in range(int(bin[j + 1]), int(bin[j + 2])):
            fbank[j, i] = (bin[j + 2] - i) / (bin[j + 2] - bin[j + 1])
    return fbank



import numbers
import warnings
from collections import namedtuple
from typing import List, Tuple

import torch
import torch.jit as jit
import torch.nn as nn
from torch import Tensor
from torch.nn import Parameter


"""
Some helper classes for writing custom TorchScript LSTMs.

Goals:
- Classes are easy to read, use, and extend
- Performance of custom LSTMs approach fused-kernel-levels of speed.

A few notes about features we could add to clean up the below code:
- Support enumerate with nn.ModuleList:
  https://github.com/pytorch/pytorch/issues/14471
- Support enumerate/zip with lists:
  https://github.com/pytorch/pytorch/issues/15952
- Support overriding of class methods:
  https://github.com/pytorch/pytorch/issues/10733
- Support passing around user-defined namedtuple types for readability
- Support slicing w/ range. It enables reversing lists easily.
  https://github.com/pytorch/pytorch/issues/10774
- Multiline type annotations. List[List[Tuple[Tensor,Tensor]]] is verbose
  https://github.com/pytorch/pytorch/pull/14922
"""


def script_lstm(
    input_size,
    hidden_size,
    num_layers,
    bias=True,
    batch_first=False,
    dropout=False,
    bidirectional=False,
):
    """Returns a ScriptModule that mimics a PyTorch native LSTM."""

    # The following are not implemented.
    assert bias
    assert not batch_first

    if bidirectional:
        stack_type = StackedLSTM2
        layer_type = BidirLSTMLayer
        dirs = 2
    elif dropout:
        stack_type = StackedLSTMWithDropout
        layer_type = LSTMLayer
        dirs = 1
    else:
        stack_type = StackedLSTM
        layer_type = LSTMLayer
        dirs = 1

    return stack_type(
        num_layers,
        layer_type,
        first_layer_args=[LSTMCell, input_size, hidden_size],
        other_layer_args=[LSTMCell, hidden_size * dirs, hidden_size],
    )


def script_lnlstm(
    input_size,
    hidden_size,
    num_layers,
    bias=True,
    batch_first=False,
    dropout=False,
    bidirectional=False,
    decompose_layernorm=False,
):
    """Returns a ScriptModule that mimics a PyTorch native LSTM."""

    # The following are not implemented.
    assert bias
    assert not batch_first
    assert not dropout

    if bidirectional:
        stack_type = StackedLSTM2
        layer_type = BidirLSTMLayer
        dirs = 2
    else:
        stack_type = StackedLSTM
        layer_type = LSTMLayer
        dirs = 1

    return stack_type(
        num_layers,
        layer_type,
        first_layer_args=[
            LayerNormLSTMCell,
            input_size,
            hidden_size,
            decompose_layernorm,
        ],
        other_layer_args=[
            LayerNormLSTMCell,
            hidden_size * dirs,
            hidden_size,
            decompose_layernorm,
        ],
    )


LSTMState = namedtuple("LSTMState", ["hx", "cx"])


def reverse(lst):
    # type: (List[Tensor]) -> List[Tensor]
    return lst[::-1]


class LSTMCell(jit.ScriptModule):
    def __init__(self, input_size, hidden_size):
        super(LSTMCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = Parameter(torch.randn(4 * hidden_size, input_size))
        self.weight_hh = Parameter(torch.randn(4 * hidden_size, hidden_size))
        self.bias_ih = Parameter(torch.randn(4 * hidden_size))
        self.bias_hh = Parameter(torch.randn(4 * hidden_size))

    @jit.script_method
    def forward(self, input, state):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        hx, cx = state
        gates = torch.mm(input, self.weight_ih.t()) + self.bias_ih + torch.mm(hx, self.weight_hh.t()) + self.bias_hh
        ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        cy = (forgetgate * cx) + (ingate * cellgate)
        hy = outgate * torch.tanh(cy)

        return hy, (hy, cy)


class LayerNorm(jit.ScriptModule):
    def __init__(self, normalized_shape):
        super(LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        normalized_shape = torch.Size(normalized_shape)

        # XXX: This is true for our LSTM / NLP use case and helps simplify code
        assert len(normalized_shape) == 1

        self.weight = Parameter(torch.ones(normalized_shape))
        self.bias = Parameter(torch.zeros(normalized_shape))
        self.normalized_shape = normalized_shape

    @jit.script_method
    def compute_layernorm_stats(self, input):
        mu = input.mean(-1, keepdim=True)
        sigma = input.std(-1, keepdim=True, unbiased=False)
        return mu, sigma

    @jit.script_method
    def forward(self, input):
        mu, sigma = self.compute_layernorm_stats(input)
        return (input - mu) / sigma * self.weight + self.bias


class LayerNormLSTMCell(jit.ScriptModule):
    def __init__(self, input_size, hidden_size, decompose_layernorm=False):
        super(LayerNormLSTMCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = Parameter(torch.randn(4 * hidden_size, input_size))
        self.weight_hh = Parameter(torch.randn(4 * hidden_size, hidden_size))
        # The layernorms provide learnable biases

        if decompose_layernorm:
            ln = LayerNorm
        else:
            ln = nn.LayerNorm

        self.layernorm_i = ln(4 * hidden_size)
        self.layernorm_h = ln(4 * hidden_size)
        self.layernorm_c = ln(hidden_size)

    @jit.script_method
    def forward(self, input, state):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        hx, cx = state
        igates = self.layernorm_i(torch.mm(input, self.weight_ih.t()))
        hgates = self.layernorm_h(torch.mm(hx, self.weight_hh.t()))
        gates = igates + hgates
        ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        cy = self.layernorm_c((forgetgate * cx) + (ingate * cellgate))
        hy = outgate * torch.tanh(cy)

        return hy, (hy, cy)


class LSTMLayer(jit.ScriptModule):
    def __init__(self, cell, *cell_args):
        super(LSTMLayer, self).__init__()
        self.cell = cell(*cell_args)

    @jit.script_method
    def forward(self, input, state):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        inputs = input.unbind(0)
        outputs = torch.jit.annotate(List[Tensor], [])
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(outputs), state


class ReverseLSTMLayer(jit.ScriptModule):
    def __init__(self, cell, *cell_args):
        super(ReverseLSTMLayer, self).__init__()
        self.cell = cell(*cell_args)

    @jit.script_method
    def forward(self, input, state):
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        inputs = reverse(input.unbind(0))
        outputs = jit.annotate(List[Tensor], [])
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(reverse(outputs)), state


class BidirLSTMLayer(jit.ScriptModule):
    __constants__ = ["directions"]

    def __init__(self, cell, *cell_args):
        super(BidirLSTMLayer, self).__init__()
        self.directions = nn.ModuleList(
            [
                LSTMLayer(cell, *cell_args),
                ReverseLSTMLayer(cell, *cell_args),
            ]
        )

    @jit.script_method
    def forward(self, input, states):
        # type: (Tensor, List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]
        # List[LSTMState]: [forward LSTMState, backward LSTMState]
        outputs = jit.annotate(List[Tensor], [])
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        for direction in self.directions:
            state = states[i]
            out, out_state = direction(input, state)
            outputs += [out]
            output_states += [out_state]
            i += 1
        return torch.cat(outputs, -1), output_states


def init_stacked_lstm(num_layers, layer, first_layer_args, other_layer_args):
    layers = [layer(*first_layer_args)] + [layer(*other_layer_args) for _ in range(num_layers - 1)]
    return nn.ModuleList(layers)


class StackedLSTM(jit.ScriptModule):
    __constants__ = ["layers"]  # Necessary for iterating through self.layers

    def __init__(self, num_layers, layer, first_layer_args, other_layer_args):
        super(StackedLSTM, self).__init__()
        self.layers = init_stacked_lstm(num_layers, layer, first_layer_args, other_layer_args)

    @jit.script_method
    def forward(self, input, states):
        # type: (Tensor, List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]
        # List[LSTMState]: One state per layer
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])
        output = input
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        for rnn_layer in self.layers:
            state = states[i]
            output, out_state = rnn_layer(output, state)
            output_states += [out_state]
            i += 1
        return output, output_states


# Differs from StackedLSTM in that its forward method takes
# List[List[Tuple[Tensor,Tensor]]]. It would be nice to subclass StackedLSTM
# except we don't support overriding script methods.
# https://github.com/pytorch/pytorch/issues/10733
class StackedLSTM2(jit.ScriptModule):
    __constants__ = ["layers"]  # Necessary for iterating through self.layers

    def __init__(self, num_layers, layer, first_layer_args, other_layer_args):
        super(StackedLSTM2, self).__init__()
        self.layers = init_stacked_lstm(num_layers, layer, first_layer_args, other_layer_args)

    @jit.script_method
    def forward(self, input, states):
        # type: (Tensor, List[List[Tuple[Tensor, Tensor]]]) -> Tuple[Tensor, List[List[Tuple[Tensor, Tensor]]]]
        # List[List[LSTMState]]: The outer list is for layers,
        #                        inner list is for directions.
        output_states = jit.annotate(List[List[Tuple[Tensor, Tensor]]], [])
        output = input
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        for rnn_layer in self.layers:
            state = states[i]
            output, out_state = rnn_layer(output, state)
            output_states += [out_state]
            i += 1
        return output, output_states


class StackedLSTMWithDropout(jit.ScriptModule):
    # Necessary for iterating through self.layers and dropout support
    __constants__ = ["layers", "num_layers"]

    def __init__(self, num_layers, layer, first_layer_args, other_layer_args):
        super(StackedLSTMWithDropout, self).__init__()
        self.layers = init_stacked_lstm(num_layers, layer, first_layer_args, other_layer_args)
        # Introduces a Dropout layer on the outputs of each LSTM layer except
        # the last layer, with dropout probability = 0.4.
        self.num_layers = num_layers

        if num_layers == 1:
            warnings.warn(
                "dropout lstm adds dropout layers after all but last "
                "recurrent layer, it expects num_layers greater than "
                "1, but got num_layers = 1"
            )

        self.dropout_layer = nn.Dropout(0.4)

    @jit.script_method
    def forward(self, input, states):
        # type: (Tensor, List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]
        # List[LSTMState]: One state per layer
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])
        output = input
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        for rnn_layer in self.layers:
            state = states[i]
            output, out_state = rnn_layer(output, state)
            # Apply the dropout layer except the last layer
            if i < self.num_layers - 1:
                output = self.dropout_layer(output)
            output_states += [out_state]
            i += 1
        return output, output_states


def flatten_states(states):
    states = list(zip(*states))
    assert len(states) == 2
    return [torch.stack(state) for state in states]


def double_flatten_states(states):
    # XXX: Can probably write this in a nicer way
    states = flatten_states([flatten_states(inner) for inner in states])
    return [hidden.view([-1] + list(hidden.shape[2:])) for hidden in states]


def script_rnn_layer(seq_len, batch, input_size, hidden_size):
    inp = torch.randn(seq_len, batch, input_size)
    state = LSTMState(torch.randn(batch, hidden_size), torch.randn(batch, hidden_size))
    rnn = LSTMLayer(LSTMCell, input_size, hidden_size)
    out, out_state = rnn(inp, state)

    # Control: pytorch native LSTM
    lstm = nn.LSTM(input_size, hidden_size, 1)
    lstm_state = LSTMState(state.hx.unsqueeze(0), state.cx.unsqueeze(0))
    for lstm_param, custom_param in zip(lstm.all_weights[0], rnn.parameters()):
        assert lstm_param.shape == custom_param.shape
        with torch.no_grad():
            lstm_param.copy_(custom_param)
    lstm_out, lstm_out_state = lstm(inp, lstm_state)

    assert (out - lstm_out).abs().max() < 1e-5
    assert (out_state[0] - lstm_out_state[0]).abs().max() < 1e-5
    assert (out_state[1] - lstm_out_state[1]).abs().max() < 1e-5


def script_stacked_rnn(seq_len, batch, input_size, hidden_size, num_layers):
    inp = torch.randn(seq_len, batch, input_size)
    states = [LSTMState(torch.randn(batch, hidden_size), torch.randn(batch, hidden_size)) for _ in range(num_layers)]
    rnn = script_lstm(input_size, hidden_size, num_layers)
    out, out_state = rnn(inp, states)
    custom_state = flatten_states(out_state)

    # Control: pytorch native LSTM
    lstm = nn.LSTM(input_size, hidden_size, num_layers)
    lstm_state = flatten_states(states)
    for layer in range(num_layers):
        custom_params = list(rnn.parameters())[4 * layer : 4 * (layer + 1)]
        for lstm_param, custom_param in zip(lstm.all_weights[layer], custom_params):
            assert lstm_param.shape == custom_param.shape
            with torch.no_grad():
                lstm_param.copy_(custom_param)
    lstm_out, lstm_out_state = lstm(inp, lstm_state)

    assert (out - lstm_out).abs().max() < 1e-5
    assert (custom_state[0] - lstm_out_state[0]).abs().max() < 1e-5
    assert (custom_state[1] - lstm_out_state[1]).abs().max() < 1e-5


def script_stacked_bidir_rnn(seq_len, batch, input_size, hidden_size, num_layers):
    inp = torch.randn(seq_len, batch, input_size)
    states = [
        [LSTMState(torch.randn(batch, hidden_size), torch.randn(batch, hidden_size)) for _ in range(2)]
        for _ in range(num_layers)
    ]
    rnn = script_lstm(input_size, hidden_size, num_layers, bidirectional=True)
    out, out_state = rnn(inp, states)
    custom_state = double_flatten_states(out_state)

    # Control: pytorch native LSTM
    lstm = nn.LSTM(input_size, hidden_size, num_layers, bidirectional=True)
    lstm_state = double_flatten_states(states)
    for layer in range(num_layers):
        for direct in range(2):
            index = 2 * layer + direct
            custom_params = list(rnn.parameters())[4 * index : 4 * index + 4]
            for lstm_param, custom_param in zip(lstm.all_weights[index], custom_params):
                assert lstm_param.shape == custom_param.shape
                with torch.no_grad():
                    lstm_param.copy_(custom_param)
    lstm_out, lstm_out_state = lstm(inp, lstm_state)

    assert (out - lstm_out).abs().max() < 1e-5
    assert (custom_state[0] - lstm_out_state[0]).abs().max() < 1e-5
    assert (custom_state[1] - lstm_out_state[1]).abs().max() < 1e-5


def script_stacked_lstm_dropout(seq_len, batch, input_size, hidden_size, num_layers):
    inp = torch.randn(seq_len, batch, input_size)
    states = [LSTMState(torch.randn(batch, hidden_size), torch.randn(batch, hidden_size)) for _ in range(num_layers)]
    rnn = script_lstm(input_size, hidden_size, num_layers, dropout=True)

    # just a smoke test
    out, out_state = rnn(inp, states)


def script_stacked_lnlstm(seq_len, batch, input_size, hidden_size, num_layers):
    print(
        f"seq_len: {seq_len}, batch: {batch}, input_size: {input_size}, hidden_size: {hidden_size}, num_layers: {num_layers}"
    )
    inp = torch.randn(seq_len, batch, input_size)
    states = [LSTMState(torch.zeros(batch, hidden_size), torch.zeros(batch, hidden_size)) for _ in range(num_layers)]
    rnn = script_lnlstm(input_size, hidden_size, num_layers)

    # just a smoke test
    out, out_state = rnn(inp, states)


if __name__ == "__main__":
    # script_rnn_layer(5, 2, 3, 7)
    # script_stacked_rnn(5, 2, 3, 7, 4)
    # script_stacked_bidir_rnn(5, 2, 3, 7, 4)
    # script_stacked_lstm_dropout(5, 2, 3, 7, 4)
    # script_stacked_lnlstm(5, 2, 3, 7, 4)
    cell = LSTMCell(3, 7)
    ipt = torch.randn(2, 3)
    states = (torch.randn(2, 7), torch.randn(2, 7))
    print(cell.graph_for(ipt, states))



import logging
from typing import Tuple

import torch
import torch.distributed as dist
from torch import Tensor


logger = logging.getLogger(__name__)


def init_accelerator(device: str = "cpu") -> Tuple[int, int]:
    """Setup the accelerator.

    Return the world size and current rank.

    Note:
        * The AudioZen library only supports CPU and NVIDIA GPUs.
        * For CPU, you can set the `device` to `cpu`. Then just perform `python script.py` and it will run on CPU.
        * For GPU, we use `torch.distributed` to run on multiple GPUs. You may set the `device` to `gpu` and run the
        script with `torchrun` command. For example, `torchrun --nproc_per_node=2 script.py` will run the script on 2
        GPUs.
        * You may found the following code snippet is very short and simple, but it's very useful for these two cases.

    Args:
        device: Device. Defaults to `cpu`. Can be `cpu` or `gpu`.
    """
    world_size, rank = 1, 0

    if device == "gpu":
        assert torch.cuda.is_available(), "CUDA device requested but not available."
        # Initialize the default distributed process group
        dist.init_process_group(backend="nccl")

        # Some PyTorch APIs will use the `torch.cuda.current_device()` to get the current device, e.g., gather(). So,
        # we need to set the current device to the rank of the current process. Otherwise, those APIs will be failed,
        # e.g., gather() will hang if the current device is not set.
        torch.cuda.set_device(dist.get_rank())

        world_size, rank = dist.get_world_size(), dist.get_rank()
    elif device == "cpu":
        pass
    else:
        raise ValueError(f"Invalid device: {device}. Please choose 'cpu' or 'gpu'.")

    # We don't use logging here because the logger may not be initialized yet.
    print(f"Accelerator initialized with device `{device}`. World size: {world_size}, Rank: {rank}")

    return world_size, rank


def get_rank() -> int:
    """Function that gets the rank number of the current process in the default process group.

    Returns:
        int: rank
    """
    if dist.is_available() and dist.is_initialized():
        return dist.get_rank()
    else:
        return 0


def get_world_size_and_rank() -> Tuple[int, int]:
    """Function that gets the current world size (aka total number of ranks) and rank number of the current process in
    the default process group.

    Return the world size and rank of the current process.
    """
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size(), dist.get_rank()
    else:
        return 1, 0


def is_rank_zero() -> bool:
    """Check if the current process is rank 0.

    Returns:
        bool: True if rank is 0, False otherwise.
    """
    return get_rank() == 0


def broadcast_tensor(tensor: Tensor, src: int = 0) -> Tensor:
    """Broadcast a tensor to all other processes.

    Args:
        tensor: The tensor to broadcast.
        src: Source rank. Defaults to 0.
    """
    if dist.is_available() and dist.is_initialized():
        # Broadcasts the tensor to all other processes in the process group.
        # Now tensor values are the same across all processes.
        dist.broadcast(tensor, src=src, async_op=False)
        return tensor
    else:
        logger.warning("Distributed environment not initialized. No broadcast will be performed.")
        return tensor


def wait_for_everyone():
    """Will stop the execution of the current process until every other process has reached that point (so this does
    nothing when the script is only run in one process).

    Note:
        After pytorch 2.5, we need to add a device_ids argument to the barrier function.

    Useful to do before saving a model. It only works if the accelerator is set to "gpu". If the accelerator is set to
    "cpu", it will do nothing.
    """
    if dist.is_available() and dist.is_initialized():
        dist.barrier(device_ids=[get_rank()])
    else:
        logger.warning("Distributed environment not initialized. No barrier will be performed.")


def gather_object(object, dst=0):
    """Gather a single object from all processes and returns it to the destination process.

    Args:
        object: The object to gather.
        dst: Destination rank. Defaults to 0.
    """
    if dist.is_available() and dist.is_initialized():
        # `object_gather_list` means a placeholder for the collected step_output from all processes.
        # `len(object_gather_list)` >= `world_size`, the extra elements will stay as `None`.
        # `len(object_gather_list)` < `world_size`, the program will raise an error.
        # The rank 0 process will gather the step_output from all processes. Other processes will not gather the
        # step_output and keep all `None`.
        output = [None for _ in range(dist.get_world_size())]
        dist.gather_object(object, output if is_rank_zero() else None, dst=dst)
        # `gather_object` returns a list of lists, so we need to flatten it
        if is_rank_zero():
            return [x for y in output for x in y]  # type: ignore
        else:
            return None
    else:
        raise RuntimeError("Distributed environment not initialized. No gather will be performed.")



import torch
import torch.nn as nn

from audiozen.models.module.custom_lstm import script_lnlstm
from audiozen.models.module.groupGRU import SharedGroupGRU
from audiozen.models.module.mogrifier_lstm import MogrifierLSTMWrapper


class SequenceModel(nn.Module):
    def __init__(
        self,
        input_size,
        output_size,
        hidden_size,
        num_layers,
        bidirectional,
        sequence_model="GRU",
        output_activate_function="Tanh",
        num_groups=4,
        mogrify_steps=5,
        dropout=0.0,
    ):
        """Wrapper of conventional sequence models.

        Args:
            input_size: input size
            output_size: when projection_size> 0, the linear layer is used for projection. Otherwise no linear layer.
            hidden_size: hidden size
            num_layers:  num layers
            bidirectional: if bidirectional
            sequence_model: LSTM, GRU, SharedGroupGRU, LayerNormLSTM
            output_activate_function: Tanh, ReLU, ReLU6, LeakyReLU, PReLU
        """
        super().__init__()
        if sequence_model == "LSTM":
            self.sequence_model = nn.LSTM(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                bidirectional=bidirectional,
                dropout=dropout,
            )
        elif sequence_model == "GRU":
            self.sequence_model = nn.GRU(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                bidirectional=bidirectional,
                dropout=dropout,
            )
        elif sequence_model == "SharedGroupGRU":
            self.sequence_model = SharedGroupGRU(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                num_groups=num_groups,
                bidirectional=bidirectional,
            )
        elif sequence_model == "LayerNormLSTM":
            self.sequence_model = script_lnlstm(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                bidirectional=bidirectional,
            )
        elif sequence_model == "MogrifierLSTM":
            assert bidirectional is False, "MogrifierLSTM does not support bidirectional"
            self.sequence_model = MogrifierLSTMWrapper(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                mogrify_steps=mogrify_steps,
            )
        else:
            raise NotImplementedError(f"Not implemented {sequence_model}")

        # Fully connected layer
        if int(output_size):
            if bidirectional:
                self.fc_output_layer = nn.Linear(hidden_size * 2, output_size)
            else:
                self.fc_output_layer = nn.Linear(hidden_size, output_size)

        # Activation function layer
        if output_activate_function:
            if output_activate_function == "Tanh":
                self.activate_function = nn.Tanh()
            elif output_activate_function == "ReLU":
                self.activate_function = nn.ReLU()
            elif output_activate_function == "ReLU6":
                self.activate_function = nn.ReLU6()
            elif output_activate_function == "LeakyReLU":
                self.activate_function = nn.LeakyReLU()
            elif output_activate_function == "PReLU":
                self.activate_function = nn.PReLU()
            else:
                raise NotImplementedError(f"Not implemented activation function {self.activate_function}")

        self.output_activate_function = output_activate_function
        self.output_size = output_size

        # only for custom_lstm and are needed to clean up
        self.sequence_model_name = sequence_model
        self.hidden_size = hidden_size
        self.num_layers = num_layers

    def forward(self, x):
        """
        Args:
            x: [B, F, T]
        Returns:
            [B, F, T]
        """
        n_dim = x.dim()
        assert n_dim in (3, 4), f"Shape is {x.shape}."

        if n_dim == 4:
            batch_size, num_channels, _, num_frames = x.shape
            x = x.reshape(batch_size, -1, num_frames)

        batch_size, _, _ = x.shape

        if self.sequence_model_name == "LayerNormLSTM":
            states = [
                (
                    torch.zeros(batch_size, self.hidden_size, device=x.device),
                    torch.zeros(batch_size, self.hidden_size, device=x.device),
                )
                for _ in range(self.num_layers)
            ]
        else:
            states = None

        x = x.permute(2, 0, 1).contiguous()  # [B, F, T] => [T, B, F]
        o, _ = self.sequence_model(x, states)  # [T, B, F] => [T, B, F]

        if self.output_size:
            o = self.fc_output_layer(o)  # [T, B, F] => [T, B, F]

        if self.output_activate_function:
            o = self.activate_function(o)

        o = o.permute(1, 2, 0).contiguous()  # [T, B, F] => [B, F, T]

        if n_dim == 4:
            o = o.reshape(batch_size, num_channels, -1, num_frames)

        return o


def _print_networks(nets: list):
    print(f"This project contains {len(nets)} networks, the number of the parameters: ")
    params_of_all_networks = 0
    for i, net in enumerate(nets, start=1):
        params_of_network = 0
        for param in net.parameters():
            params_of_network += param.numel()

        print(f"\tNetwork {i}: {params_of_network / 1e6} million.")
        params_of_all_networks += params_of_network

    print(f"The amount of parameters in the project is {params_of_all_networks / 1e6} million.")


if __name__ == "__main__":
    # with torch.no_grad():

    #
    #     ipt = torch.rand(batch, input_size, seq_len)
    #     # states = [LSTMState(torch.zeros(batch, hidden_size),
    #     #                     torch.zeros(batch, hidden_size))
    #     #           for _ in range(num_layers)]
    #
    #     model = SequenceModel(
    #         input_size=input_size,
    #         output_size=2,
    #         hidden_size=512,
    #         bidirectional=False,
    #         num_layers=3,
    #         sequence_model="LayerNormLSTM"
    #     )
    #
    #     start = datetime.datetime.now()
    #     opt = model(ipt)
    #     end = datetime.datetime.now()
    #     print(f"{end - start}")
    #     _print_networks([model, ])

    batch_size = 2
    hidden_size = 512
    output_size = 258
    num_layers = 3
    seq_len = 1000
    input_size = 257
    ipt = torch.rand(batch_size, input_size, seq_len)
    ipt = ipt.to("cuda:0")

    model = SequenceModel(
        input_size=input_size,
        output_size=output_size,
        hidden_size=hidden_size,
        bidirectional=False,
        num_layers=num_layers,
        sequence_model="LSTM",
    )

    # model.to("cuda:0")
    #
    # opt = model(ipt)
    # print(opt.shape)



import math

import numpy as np
import torch


EPS = torch.as_tensor(torch.finfo(torch.get_default_dtype()).eps)
PI = math.pi


def db2mag(fdB):
    """Converts dB to magnitude."""
    return 10 ** (fdB / 20)


class STFTTorch:
    """Class used to simplify handling of STFT & iSTFT."""

    def __init__(
        self,
        frame_length=64,
        overlap_length=48,
        window=torch.hann_window,
        fs=16000,
    ):
        self.num_bins = int((frame_length / 2) + 1)
        self.fs = fs
        self.frame_length = frame_length
        self.overlap_length = overlap_length
        self.window = window(self.frame_length)
        self.shift_length = self.frame_length - self.overlap_length

        self.params = {
            "frame_length": frame_length,
            "overlap_length": overlap_length,
            "window": window,
        }

    def get_stft(self, wave):
        return torch.stft(
            wave,
            window=self.window.to(device=wave.device),
            n_fft=self.frame_length,
            hop_length=self.shift_length,
            win_length=self.frame_length,
            normalized=False,
            center=True,
            pad_mode="constant",
        )

    def get_istft(self, stft, length=None):
        return torch.istft(
            stft,
            window=self.window.to(device=stft.device),
            n_fft=self.frame_length,
            hop_length=self.shift_length,
            win_length=self.frame_length,
            normalized=False,
            center=True,
            length=length,
        )


def complex_tensor_exponential(t):
    """Return complex exponential of input with shape ...

    x 2
    """
    exp_real = torch.exp(t[..., 0])
    real = exp_real * torch.cos(t[..., 1])
    imag = exp_real * torch.sin(t[..., 1])
    return torch.stack([real, imag], dim=-1)


def vector_to_Hermitian(vec):
    """This function constructs a ...

    x N x N x 2-dim. Hermitian matrix from a vector of N**2 independent real-valued components
    broadcasting enabled
    input:
        vec: ... x N**2
    output:
        mat: ... x N x N x 2
    """
    N = int(np.sqrt(vec.shape[-1]))
    mat = torch.zeros(size=vec.shape[:-1] + (N, N, 2), device=vec.device)

    # real component
    triu = np.triu_indices(N, 0)
    triu2 = np.triu_indices(N, 1)  # above main diagonal
    tril = (triu2[1], triu2[0])  # below main diagonal; for symmetry
    mat[(...,) + triu + (np.zeros(triu[0].shape[0]),)] = vec[..., : triu[0].shape[0]]
    start = triu[0].shape[0]
    mat[(...,) + tril + (np.zeros(tril[0].shape[0]),)] = mat[(...,) + triu2 + (np.zeros(triu2[0].shape[0]),)]

    # imaginary component
    mat[(...,) + triu2 + (np.ones(triu2[0].shape[0]),)] = vec[..., start : start + triu2[0].shape[0]]
    mat[(...,) + tril + (np.ones(tril[0].shape[0]),)] = -mat[(...,) + triu2 + (np.ones(triu2[0].shape[0]),)]

    return mat


def complex_tensor_hermitian(t: torch.Tensor) -> torch.Tensor:
    """Return Hermitian of t.

    Args:
        t (torch.Tensor): ... x N x N x 2

    Returns:
        torch.Tensor: Hermitian of t; ... x N x N x 2
    """
    return complex_tensor_conj(t.transpose(-3, -2))


def get_mvdr(gammax, Phi):
    """Compute conventional MPDR/MVDR filter :param gammax:

    :param Phi:
    :return: filter coefficients filter
    computes filter coefficients;
    gammax: KxMx2, Phi: KxMxMx2
    """
    b = complex_solve_matrix_vector(Phi, gammax)
    denom = complex_tensor_inner_product(gammax, b)
    return complex_tensor_division(b, denom[..., None, :] + EPS)


def complex_solve_matrix_vector(A, b):
    """Solves a complex system of linear equations 07.10.19: validated against numpy."""
    A_big = torch.cat(
        (
            torch.cat((A[..., 0], -A[..., 1]), dim=-1),
            torch.cat((A[..., 1], A[..., 0]), dim=-1),
        ),
        dim=-2,
    )
    b_big = torch.cat((b[..., 0], b[..., 1]), dim=-1)
    x_big = torch.solve(b_big[..., None], A_big)[0][..., 0]  # ignore singleton dimension
    length = int(x_big.shape[-1] / 2)
    return torch.stack((x_big[..., :length], x_big[..., length:]), dim=-1)


def complex_solve_matrix_matrix(mat1, mat2):
    """
    solve system of linear equations specified by
    mat1 X = mat2 <-> X = mat1^-1 mat2
    Args:
        mat1 ([torch.tensor]): ... m x n x 2 (real and imaginary components in last dimension)
        mat2 ([torch.tensor]): ... n x o x 2

    Returns:
        [torch.tensor]: X = mat1^-1 mat2; ... x m x o
    """
    n = mat1.shape[-2]
    mat1_big = torch.cat(
        (
            torch.cat((mat1[..., 0], -mat1[..., 1]), dim=-1),
            torch.cat((mat1[..., 1], mat1[..., 0]), dim=-1),
        ),
        dim=-2,
    )
    mat2_big = torch.cat((mat2[..., 0], mat2[..., 1]), dim=-2)
    solution_big = torch.solve(mat2_big, mat1_big)[0]
    return torch.stack((solution_big[..., :n, :], solution_big[..., n:, :]), dim=-1)


def filter_(mWeights, mSTFTNoisyAdj):
    """
    performs filtering on STFT using mWeights
    mWeights: batch_size x K x time_steps x AdjFrames
    """
    return complex_tensor_inner_product(mWeights, mSTFTNoisyAdj)


def filter_minimum_gain_like(G_min, w, y, alpha=None, k=10.0):
    """Approximate a minimum gain operation as.

    speech_estimate = alpha w^H y + (1 - alpha) G_min Y,
    where alpha = 1 / (1 + exp(-2 k x)), x = w^H y - G_min Y
    inputs:
        - G_min:    minimum gain, float
        - w:        complex-valued filter coefficients, ... x L x N x 2
        - y:        buffered and stacked input, ... x L x N x 2
        - k:        scaling in tanh-like function
        - alpha:    mixing factor
    outputs:
        - minimum gain-filtered output
        - (optional) alpha
    """
    filtered_input = complex_tensor_inner_product(w, y)
    Y = y[..., -1, :]
    return minimum_gain_like(G_min, Y, filtered_input, alpha, k)


def minimum_gain_like(G_min, Y, filtered_input, alpha=None, k=10.0):
    if alpha is None:
        alpha = 1.0 / (1.0 + torch.exp(-2 * k * (complex_tensor_abs(filtered_input) - complex_tensor_abs(G_min * Y))))
        alpha = alpha[..., None]
        return_alpha = True
    else:
        return_alpha = False
    output = alpha * filtered_input + (1 - alpha) * G_min * Y
    if return_alpha:
        return output, alpha
    else:
        return output


@torch.jit.script
def complex_tensor_hadamard_product(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """Perform multiplication of two complex tensors a and b, with real and imag components in last dimensions :param
    a:

    :param b:
    :return:
    """
    return torch.stack(
        (
            a[..., 0] * b[..., 0] - a[..., 1] * b[..., 1],
            a[..., 0] * b[..., 1] + a[..., 1] * b[..., 0],
        ),
        dim=-1,
    )


@torch.jit.script
def safe_sqrt(tensor, eps=EPS):
    """Safer version of sqrt, adding EPS to all values to avoid nan :param input:

    :param kwargs:
    :return:
    """
    return torch.sqrt(tensor + eps)


@torch.jit.script
def complex_tensor_conj(t: torch.Tensor) -> torch.Tensor:
    """Return complex conjugate, assuming that real and imag parts lie in last dimension :param mat:

    :return:
    """
    return torch.stack((t[..., 0], -t[..., 1]), dim=-1)


@torch.jit.script
def complex_tensor_matrix_vector_product(matrix: torch.Tensor, vector: torch.Tensor) -> torch.Tensor:
    """Perform multiplication of complex tensor matrix and vector with real and imag components in last dimensions."""
    return torch.stack(
        [
            matrix[..., 0] @ vector[..., 0][..., None] - matrix[..., 1] @ vector[..., 1][..., None],
            matrix[..., 0] @ vector[..., 1][..., None] + matrix[..., 1] @ vector[..., 0][..., None],
        ],
        dim=-1,
    )[..., 0, :]


@torch.jit.script
def complex_tensor_matrix_matrix_product(mat1: torch.Tensor, mat2: torch.Tensor) -> torch.Tensor:
    """matrix multiplication of complex matrices mat1 and mat2
    Args:
        mat1 ([tensor]): ... x M x N x 2 complex tensor
        mat2 ([tensor]): ... x N x O x 2 complex tensor
    """
    return torch.stack(
        [
            mat1[..., 0] @ mat2[..., 0] - mat1[..., 1] @ mat2[..., 1],
            mat1[..., 0] @ mat2[..., 1] + mat1[..., 1] @ mat2[..., 0],
        ],
        dim=-1,
    )


def trace(mat: torch.Tensor) -> torch.Tensor:
    """Returns the trace of mat, taken over the last two dimensions :param mat:

    :return:
    """
    return torch.einsum("...ii->...", mat)


@torch.jit.script
def complex_tensor_abs(tIn: torch.Tensor) -> torch.Tensor:
    """
    returns the absolute value of tIn
    tIn: ... x 2, with real and imaginary component on last axis
    """
    return safe_sqrt(tIn[..., 0] ** 2 + tIn[..., 1] ** 2)


@torch.jit.script
def tik_reg(mat: torch.Tensor, reg: float = 0.001) -> torch.Tensor:
    """
    performs Tikhonov regularization
    only modifies real part
    mat: ... x iAdj x 2
    """
    iAdj = mat.shape[-2]  # number of adjacent frames
    temp = ((reg * trace(complex_tensor_abs(mat))) / iAdj)[..., None, None] * torch.eye(iAdj, device=mat.device)[
        None, None, ...
    ]
    return mat + torch.stack([temp, temp.new_zeros(size=(1,)).expand(temp.shape)], dim=-1)


@torch.jit.script
def complex_tensor_division(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """Perform division a/b, with a and b complex tensors, with real and imaginary components sitting in last dim.

    :param a:
    :param b:
    :return:
    """
    factor = 1.0 / (b[..., 0] ** 2 + b[..., 1] ** 2)
    result = torch.stack(
        (
            a[..., 0] * b[..., 0] + a[..., 1] * b[..., 1],
            a[..., 1] * b[..., 0] - a[..., 0] * b[..., 1],
        ),
        dim=-1,
    )
    return factor[..., None] * result


@torch.jit.script
def complex_tensor_inner_product(x, y):
    """Perform inner product of two complex-valued vectors."""
    xHy = torch.empty(size=x[..., 0, :].shape, device=x.device)
    xHy[..., 0] = torch.sum(x[..., 0] * y[..., 0] + x[..., 1] * y[..., 1], dim=-1)
    xHy[..., 1] = torch.sum(x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0], dim=-1)
    return xHy


class Bunch(object):
    """Transforms adict into an object such that adict values can be accessed using "." notation."""

    def __init__(self, adict) -> None:
        self.__dict__.update(adict)



import math
import shutil
import time
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import librosa
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import DistributedDataParallelKwargs, GradientAccumulationPlugin, set_seed
from torch.utils.data import DataLoader, Dataset, RandomSampler
from torchinfo import summary
from tqdm.auto import tqdm

from audiozen.acoustics.audio_feature import istft, stft
from audiozen.debug_utils import DebugUnderflowOverflow
from audiozen.logger import TensorboardLogger
from audiozen.optimization import get_constant_schedule_with_warmup, get_linear_schedule_with_warmup
from audiozen.trainer_args import TrainingArgs
from audiozen.trainer_utils import TrainerState, collect_env, seed_worker
from audiozen.utils import prepare_empty_dir


logger = get_logger(__name__)


class Trainer:
    def __init__(
        self,
        model: nn.Module,
        args: TrainingArgs,
        data_collator: Optional[Any] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
        optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
    ):
        """Initialize the trainer.

        Args:
            model: The model to train, evaluate or use for predictions.
            args: The training arguments. It is a object of `audiozen.trainer_args.TrainingArgs`.
            data_collator: The data collator to use for training and evaluation. Defaults to None.
            train_dataset: The training dataset. Defaults to None.
            eval_dataset: The evaluation dataset. Defaults to None. It can be a single dataset or a dictionary of datasets.
            optimizers: The optimizer and the learning rate scheduler. Defaults to (None, None).

        Important attributes:
            **is_in_train**: Whether or not a model is currently running `train` (e.g. when `evaluate` is called while in `train`)
        """
        self.is_in_train = False
        self.args = args

        # Accelerator should be created as early as possible
        self.create_accelerator()

        # Set random seed using the method from accelerate
        set_seed(self.args.seed, device_specific=True)

        # Create model
        # We are able to `prepare` the model directly in the `__init__` method.
        # However, there is a bug in the `prepare` method. We must prepare model and optimizer simultaneously. (2024-03-31)
        self.model = model

        # Optimizers
        self.optimizer, self.lr_scheduler = optimizers

        # Datasets
        self.data_collator = data_collator
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset

        # Dataloaders. They will be created in the `train` and `evaluate` methods.
        self.train_dataloader = None
        self.eval_dataloaders = None

        # Setup directories
        self._setup_exp_paths(output_dir=self.args.output_dir)

        # Acoustic args
        self._setup_acoustic_args()

        # Trainer states
        self.state = TrainerState(greater_is_better=self.args.greater_is_better)

        # Register accelerate objects
        self.accelerator.register_for_checkpointing(self.state)

        # Pandas settings for better display
        pd.set_option("display.float_format", lambda x: "%.3f" % x)

        if self.accelerator.is_local_main_process:
            prepare_empty_dir(
                [self.checkpoints_dir, self.tb_log_dir, self.enhanced_dir, self.metrics_dir],
                resume=self.args.resume_from_checkpoint != "no",
            )

            logger.info(f"\nEnvironment information:\n{collect_env()}")

            self.writer = TensorboardLogger(self.tb_log_dir.as_posix())
            self.writer.log_config(self.args.to_dict())

            # Backup of project code
            # shutil.copytree(src=self.source_code_dir.as_posix(), dst=self.source_code_backup_dir.as_posix())
            # logger.info(f"Project code is backed up to {self.source_code_backup_dir.as_posix()}.")

            # Model summary
            logger.info(f"\n{summary(self.model, verbose=0)}")

        self.accelerator.wait_for_everyone()

    def create_accelerator(self):
        grad_acc_kwargs = {"num_steps": self.args.gradient_accumulation_steps}
        grad_acc_kwargs["sync_with_dataloader"] = False
        gradient_accumulation_plugin = GradientAccumulationPlugin(**grad_acc_kwargs)

        accelerator = Accelerator(
            gradient_accumulation_plugin=gradient_accumulation_plugin,
            kwargs_handlers=[
                DistributedDataParallelKwargs(find_unused_parameters=self.args.ddp_find_unused_parameters)
            ],
        )

        self.rank = accelerator.device
        self.device = accelerator.device  # alias of rank
        self.process_index = accelerator.process_index
        self.num_devices = accelerator.num_processes
        self.accelerator = accelerator

    @staticmethod
    def get_optimizer_cls_and_kwargs(args: TrainingArgs) -> Tuple[torch.optim.Optimizer, Dict]:
        """Returns the optimizer class and optimizer parameters based on the training arguments."""
        optimizer_kwargs = {"lr": args.learning_rate}

        adam_kwargs = {"betas": (args.adam_beta1, args.adam_beta2), "eps": args.adam_epsilon}

        if args.optim == "adamw":
            optimizer_cls = torch.optim.AdamW
            optimizer_kwargs |= adam_kwargs
        elif args.optim == "adam":
            optimizer_cls = torch.optim.Adam
            optimizer_kwargs |= adam_kwargs
        else:
            raise ValueError(f"Unknown optimizer: {args.optim}")

        return optimizer_cls, optimizer_kwargs

    def create_optimizer(self):
        """Setup the optimizer."""
        if self.optimizer is None:
            optimizer_cls, optimizer_kwargs = self.get_optimizer_cls_and_kwargs(self.args)
            self.optimizer = optimizer_cls(self.model.parameters(), **optimizer_kwargs)

        return self.optimizer

    @staticmethod
    def get_warmup_steps(warmup_steps, max_steps, warmup_ratio):
        if warmup_steps > 0:
            logger.info(f"warmup_steps={warmup_steps}. warmup_ratio will be ignored.")
            return warmup_steps
        else:
            return math.ceil(max_steps * warmup_ratio)

    def create_warmup_scheduler(self, optimizer, scheduler_name, max_steps: int):
        num_warmup_steps = self.get_warmup_steps(self.args.warmup_steps, max_steps, self.args.warmup_ratio)
        if scheduler_name == "constant_schedule_with_warmup":
            return get_constant_schedule_with_warmup(optimizer=optimizer, num_warmup_steps=num_warmup_steps)
        elif scheduler_name == "linear_schedule_with_warmup":
            return get_linear_schedule_with_warmup(
                optimizer=optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=max_steps
            )
        else:
            raise ValueError(f"Invalid scheduler name: {scheduler_name}")

    def create_scheduler(self, max_steps: int):
        """Create schedulers.

        You can override this method to create your own schedulers. For example, in GAN training, you may want to
        create two schedulers for the generator and the discriminator.

        Args:
            max_steps: the maximum number of steps to train.
        """
        self.lr_scheduler = self.create_warmup_scheduler(
            optimizer=self.optimizer, scheduler_name=self.args.lr_scheduler_type, max_steps=max_steps
        )
        self.lr_scheduler = self.accelerator.prepare(self.lr_scheduler)

    def create_optimizer_and_scheduler(self, num_training_steps: int):
        """Setup the optimizer and the learning rate scheduler."""
        self.create_optimizer()
        self.create_scheduler(num_training_steps)

    def _run_early_stop_check(self, score: float):
        should_stop = False

        if self._check_improvement(score, save_max_score=self.args.greater_is_better):
            self.state.best_score = score
            self.state.best_score_epoch = self.state.epochs_trained
            self._save_checkpoint(self.state.epochs_trained, is_best_epoch=True)
            self.state.early_stopping_patience_counter = 0
            logger.info(f"Found new best score: {score:.4f}, saving checkpoint...")
        else:
            logger.info(
                f"Score did not improve from {self.state.best_score:.4f} at epoch {self.state.best_score_epoch}."
            )
            self.state.early_stopping_patience_counter += 1
            logger.info(
                f"Early stopping counter: {self.state.early_stopping_patience_counter} out of {self.args.early_stopping_patience}"
            )

            if self.state.early_stopping_patience_counter >= self.args.early_stopping_patience:
                logger.info("Early stopping triggered, stopping training...")
                should_stop = True

        return should_stop

    def _setup_acoustic_args(self):
        """Setup acoustic arguments."""
        n_fft = self.args.acoustic_n_fft
        hop_length = self.args.acoustic_hop_length
        win_length = self.args.acoustic_win_length
        sr = self.args.acoustic_sr

        # Support for torch and librosa stft
        self.torch_stft = partial(stft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.torch_istft = partial(istft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.librosa_stft = partial(librosa.stft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.librosa_istft = partial(librosa.istft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)

        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        self.sr = sr

    @staticmethod
    def _get_time_now():
        return time.strftime("%Y_%m_%d--%H_%M_%S")

    def _setup_exp_paths(self, output_dir: str):
        """Set the paths for the experiment.

        Args:
            output_dir: the root directory to save all experiments.

        Example:
            - output_dir: /home/xhao/exp/fullsubnet_lr_0.1
            - checkpoints_dir: /home/xhao/exp/fullsubnet_lr_0.1/checkpoints
            - tb_log_dir: /home/xhao/exp/fullsubnet_lr_0.1/tb_log
            - enhanced_dir: /home/xhao/exp/fullsubnet_lr_0.1/enhanced
            - metrics_dir: /home/xhao/exp/fullsubnet_lr_0.1/metrics
            - source_code_backup_dir: /home/xhao/exp/fullsubnet_lr_0.1/source_code__YYYY_MM_DD__HH_MM_SS
            - model_args_path: /home/xhao/exp/fullsubnet_lr_0.1/model_args__YYYY_MM_DD__HH_MM_SS.yaml
        """
        time_now = self._get_time_now()  # returns a timestamp string

        self.output_dir = Path(output_dir).expanduser().absolute()
        self.checkpoints_dir = self.output_dir / "checkpoints"
        self.tb_log_dir = self.output_dir / "tb_log"
        self.enhanced_dir = self.output_dir / "enhanced"
        self.metrics_dir = self.output_dir / "metrics"

        # Each run will have a unique source code, config, and log file.
        self.source_code_dir = Path(__file__).expanduser().absolute().parent.parent.parent
        self.source_code_backup_dir = self.output_dir / f"source_code__{time_now}"
        self.model_args_path = self.output_dir / f"model_args__{time_now}.yaml"
        self.loss_log_path = self.output_dir / "loss.csv"

    def _find_latest_ckpt_path(self):
        """Find the latest checkpoint path."""
        # Pick up all checkpoints with the format `epoch_*`
        checkpoints = sorted(self.checkpoints_dir.glob("epoch_" + ("[0-9]" * 4)))

        # Remove files that is not a checkpoint
        checkpoints = [ckpt for ckpt in checkpoints if ckpt.is_dir()]

        if len(checkpoints) == 0:
            raise FileNotFoundError(f"No checkpoints found in {self.checkpoints_dir.as_posix()}.")

        # Pick up the latest checkpoint
        ckpt_path = checkpoints[-1]

        return ckpt_path

    def _load_checkpoint(self, ckpt: str):
        """Load a specific checkpoint.

        Args:
            ckpt_path: "no", "best", "latest", or a path to a checkpoint file.
        """
        if not isinstance(ckpt, str):
            raise ValueError(f"Invalid checkpoint specifier: {ckpt}.")

        # Lower case the checkpoint string
        ckpt = ckpt.lower()

        # It can be "no", "best", "latest", or a path to a checkpoint file.
        if ckpt == "no":
            logger.info("Will start from scratch (no checkpoint will be loaded).")
            return
        elif ckpt == "best":
            ckpt_path = self.checkpoints_dir / "best"
        elif ckpt == "latest":
            ckpt_path = self._find_latest_ckpt_path()
        else:
            ckpt_path = Path(ckpt).expanduser().absolute()

        if not ckpt_path.exists():
            raise FileNotFoundError(f"Checkpoint {ckpt_path.as_posix()} not found.")

        self.accelerator.load_state(ckpt_path, map_location="cpu")

        logger.info(f"Checkpoint on epoch {self.state.epochs_trained} is loaded.")

    def _save_checkpoint(self, epoch, is_best_epoch, safe_serialization=False):
        """Save checkpoint.

        Args:
            epoch: the current epoch.
            is_best_epoch: whether the current epoch is the best epoch.
        """
        # Save checkpoint
        if is_best_epoch:
            self.accelerator.save_state(self.checkpoints_dir / "best", safe_serialization=safe_serialization)
        else:
            # Regular checkpoint
            ckpt_path = self.checkpoints_dir / f"epoch_{str(epoch).zfill(4)}"
            self.accelerator.save_state(ckpt_path.as_posix(), safe_serialization=safe_serialization)

        # Find all regular checkpoints and only keep the latest `args.save_total_limit` regular checkpoints
        checkpoints = sorted(self.checkpoints_dir.glob("epoch_*"))

        if epoch < len(checkpoints):
            logger.warning(
                f"Current epoch is {epoch}, but found {len(checkpoints)} checkpoints. "
                f"This may be caused by you running the same experiment multiple times. "
                f"Recommend to run the experiment with a different `output_dir`."
            )

        if len(checkpoints) > self.args.save_total_limit:
            logger.info(
                f"Found {len(checkpoints)} checkpoints, only keeping the latest {self.args.save_total_limit} checkpoints."
            )
            for checkpoint_dir in checkpoints[: -self.args.save_total_limit]:
                shutil.rmtree(checkpoint_dir.as_posix())
                logger.info(f"Checkpoint {checkpoint_dir.as_posix()} is removed.")

    def set_models_to_train_mode(self):
        """Set models to train mode.

        You can override this method to set your own models to train mode. For example, in GAN training, you may want
        to set the generator and the discriminator to train mode.
        """
        self.model.train()

    def set_models_to_eval_mode(self):
        self.model.eval()

    def lr_scheduler_step(self):
        """Step the lr scheduler.

        You can override this method to step your own lr scheduler. For example, in GAN training, you may want to step
        the lr scheduler of the generator and the discriminator.
        """
        self.lr_scheduler.step(self.state.steps_trained)

    def create_bar_desc(self, loss_dict: Dict[str, float]):
        bar_desc = ""
        for k, v in loss_dict.items():
            bar_desc += f"{k}: {(v):.4f}, "
        bar_desc += f"lr: {self.lr_scheduler.get_last_lr()[-1]:.10f}"

        if self.args.plot_lr:
            self.writer.add_scalar("Train_Step/lr", self.lr_scheduler.get_last_lr()[-1], self.state.steps_trained)

        return bar_desc

    def get_train_dataloader(self) -> DataLoader:
        """Create the training dataloader.

        Returns:
            train_dataloader: the training dataloader.
        """
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a `train_dataset`.")

        dataloader_params = {
            "batch_size": self.args.per_device_train_batch_size,
            "collate_fn": self.data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "drop_last": self.args.dataloader_drop_last,
            "prefetch_factor": self.args.dataloader_prefetch_factor,
            "sampler": RandomSampler(self.train_dataset),
            "worker_init_fn": seed_worker,
        }

        return self.accelerator.prepare(DataLoader(self.train_dataset, **dataloader_params))

    def get_eval_dataloaders(self) -> Dict[str, DataLoader]:
        """Create the evaluation dataloaders.

        If the eval_dataset is a single dataset, it will be converted to a dictionary with the key "default".

        Returns:
            eval_dataloaders: the evaluation dataloaders.
        """
        if self.eval_dataset is None:
            raise ValueError("Trainer: evaluation requires a `eval_dataset`.")

        eval_dataset = self.eval_dataset
        data_collator = self.data_collator

        dataloader_params = {
            "batch_size": self.args.eval_batch_size,
            "collate_fn": data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "drop_last": False,
            "prefetch_factor": self.args.dataloader_prefetch_factor,
        }

        if not isinstance(eval_dataset, dict):
            eval_dataset = {"default": eval_dataset}

        eval_dataloaders = {}
        for key, dataset in eval_dataset.items():
            eval_dataloaders[key] = self.accelerator.prepare(DataLoader(dataset, **dataloader_params))

        return eval_dataloaders

    def train(self):
        """Main training entry point.

        Args:
            train_dataloader: the dataloader to train.
            evaluation_dataloaders: the dataloader(s) to validate.

        Notes:
            You are responsible for calling ``.backward()``, ``.step()``, ``clip_grad_norm_()``, and ``.zero_grad()``
            in your implementation of `training_step()`. Accelerate will automatically handle the gradient accumulation
            for you. It means that in gradient accumulation, the step() of optimizer and scheduler is called only when
            gradient_accumulation_steps is reached.

            The training step is implemented as follows:

            .. code-block:: python

                    self.optimizer.zero_grad()
                    loss = training_step(batch, batch_idx)
                    self.accelerator.backward(loss)
                    self.optimizer.step()

                    return {"loss": loss}
        """
        self.is_in_train = True

        if self.args.debug:
            logger.info("Debug mode is on")
            DebugUnderflowOverflow(self.model)

        self.accelerator.free_memory()

        # Create dataloaders
        train_dataloader = self.get_train_dataloader()

        # Training control variables
        len_dataloader = len(train_dataloader)
        num_update_steps_per_epoch = len_dataloader // self.args.gradient_accumulation_steps
        num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
        if self.args.max_steps > 0:
            # max_steps is set, so we need to calculate the number of epochs based on max_steps
            max_steps = self.args.max_steps
            num_train_epochs = self.args.max_steps // num_update_steps_per_epoch + int(
                self.args.max_steps % num_update_steps_per_epoch > 0
            )
        else:
            max_steps = math.ceil(self.args.num_train_epochs * num_update_steps_per_epoch)
            num_train_epochs = self.args.num_train_epochs

        # Create optimizer and scheduler
        self.create_optimizer_and_scheduler(num_training_steps=max_steps)
        self.model, self.optimizer, self.lr_scheduler = self.accelerator.prepare(
            self.model, self.optimizer, self.lr_scheduler
        )

        # Load potential model checkpoint
        resume_from_checkpoint = self.args.resume_from_checkpoint
        # Resume from the latest checkpoint
        if isinstance(resume_from_checkpoint, bool):
            if resume_from_checkpoint:
                self._load_checkpoint(ckpt="latest")
            else:
                logger.info("Training from scratch.")
        # Fine-tune from a specific checkpoint
        elif isinstance(resume_from_checkpoint, str):
            self._load_checkpoint(ckpt=resume_from_checkpoint)
        else:
            raise ValueError(f"Invalid ``resume_from_checkpoint``: {resume_from_checkpoint}")

        # Train!
        logger.info("***** Running training *****")
        logger.info(f"  Num Epochs = {num_train_epochs:,}")
        logger.info(f"  `steps_per_epoch` = {num_update_steps_per_epoch:,}")
        logger.info(f"  Instantaneous batch size per device = {self.args.per_device_train_batch_size:,}")
        logger.info(f"  Gradient Accumulation steps = {self.args.gradient_accumulation_steps}")
        logger.info(f"  Total optimization steps = {max_steps:,}")

        for epoch in range(self.state.epochs_trained + 1, num_train_epochs + 1):
            logger.info(f"{'=' * 9} Epoch {epoch} out of {num_train_epochs} {'=' * 9}")
            logger.info("Begin training...")

            self.set_models_to_train_mode()

            training_epoch_output = []

            # the iter number of progress bar increments by 1 by default whether gradient accumulation is used or not.
            # but we update the description of the progress bar only when the gradients are synchronized across all processes.
            dataloader_bar = tqdm(
                train_dataloader,
                desc="",
                dynamic_ncols=True,
                bar_format="{l_bar}{r_bar}",
                colour="green",
                disable=not self.accelerator.is_local_main_process,
                position=0,
                leave=True,
            )

            for batch_idx, batch in enumerate(dataloader_bar):
                # accumulate() will automatically skip synchronization if applicable loss is linearly scaled with the optimizer.grad
                # accumulate() will automatically divide the loss in backward by the number of gradient accumulation steps
                # However, it won't return this loss, so we need to manually divide the loss by the number of gradient accumulation steps.
                with self.accelerator.accumulate(self.model):
                    # You are responsible for calling `.backward()`, `.step()`, and `.zero_grad()` in your implementation
                    loss_dict = self.training_step(batch, batch_idx)

                    # I guess we don't need to divide the loss by the number of gradient accumulation steps here
                    # for visualization, we just plot the mean of mean of the loss of each batch
                    training_epoch_output.append(loss_dict)

                    # If `sync_gradients` is True, the gradients are currently being synced across all processes.
                    # It means that the current step we have finished the cumulative gradient accumulation.
                    if self.accelerator.sync_gradients and self.accelerator.is_local_main_process:
                        bar_desc = self.create_bar_desc(loss_dict)
                        dataloader_bar.set_description_str(bar_desc)

                    if not self.accelerator.optimizer_step_was_skipped:
                        # We can put the scheduler.step() into the training_step() function. However, it has **too much
                        # details should be considered**. It's better to put it here and add some comments.
                        #
                        # 1. every process lr_scheduler step N times, where N is the number of processes. We need to multiply the number of steps by the number of processes before constructing the
                        # scheduler to make sure it behaves as we expect it to do. https://github.com/huggingface/accelerate/issues/1398
                        #
                        # 2. For AMP, if the gradients are `nan` or `inf` skip the update step, we should call the
                        # `scheduler.step()` after checking `self.accelerator.optimizer_step_was_skipped`.
                        # Otherwise, the scheduler.step() will be called even if the optimizer step is skipped.
                        self.lr_scheduler_step()

                self.state.steps_trained += 1
            self.state.epochs_trained += 1
            self.training_epoch_end(training_epoch_output)

            # Should save, evaluate, and early stop?
            if self.accelerator.is_local_main_process and epoch % self.args.save_epoch_interval == 0:
                self._save_checkpoint(epoch, is_best_epoch=False)

            if epoch % self.args.eval_epoch_interval == 0:
                with torch.no_grad():
                    logger.info("Training finished, begin evaluation...")
                    score = self.evaluate()

                    if self.accelerator.is_local_main_process:
                        should_stop = self._run_early_stop_check(score)
                        if should_stop:
                            self.accelerator.set_trigger()

                    logger.info("evaluation finished.")

            self.accelerator.wait_for_everyone()
            # If any process triggers early stopping, stop training
            if self.accelerator.check_trigger():
                break

    @torch.no_grad()
    def evaluate(self):
        """Run evaluation (validation) and returns metrics.

        Returns:
            score: The representative score of the evaluation.
        """
        logger.info("Begin evaluation...")

        # get_eval_dataloaders() will be called multiple times, so it will prepare the dataloaders multiple times.
        # As a result, it will save multiple copies of the dataloaders in the Accelerate object.
        # Here, we will make sure that the dataloaders are only prepared once.
        if self.eval_dataloaders is None:
            self.eval_dataloaders = self.get_eval_dataloaders()

        # Whether the evaluation is running during the training or running as a standalone evaluation.
        # If the evaluation is running as a standalone evaluation, the model should be loaded from a checkpoint.
        if not self.is_in_train:
            self.model = self.accelerator.prepare(self.model)
            self._load_checkpoint(ckpt=self.args.resume_from_checkpoint)

        evaluation_output = self.evaluation_loop(description="evaluate", gather_step_output=True)

        logger.info("Evaluation finished, begin hook `evaluate_epoch_end`...")
        if self.accelerator.is_local_main_process:
            # only the main process will run evaluation_epoch_end
            score = self.evaluation_epoch_end(evaluation_output)
            return score
        else:
            return None

    @torch.no_grad()
    def predict(self):
        """Run prediction.

        In the predict mode, the model will be loaded from a checkpoint and the evaluation_loop will be called.
        However, the evaluation_loop will not gather any step_output from all processes.
        """
        logger.info("Begin predicting...")

        # In the predict mode, get_eval_dataloaders() will be called only once.
        self.eval_dataloaders = self.get_eval_dataloaders()

        self.model = self.accelerator.prepare(self.model)
        self._load_checkpoint(ckpt=self.args.resume_from_checkpoint)

        # In the predict mode, we don't need to gather the step_output from all processes.
        self.evaluation_loop(description="predict", gather_step_output=False)

        logger.info("Prediction finished.")

    @torch.no_grad()
    def evaluation_loop(self, description: str, gather_step_output: bool = False):
        """Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`."""
        args = self.args

        self.set_models_to_eval_mode()

        logger.info(f"***** Running {description} *****")
        logger.info(f"  Batch size = {args.eval_batch_size}")

        evaluation_output = {}
        for dl_idx, (dl_id, dataloader) in enumerate(self.eval_dataloaders.items()):
            dataloader_output = []
            for batch_idx, batch in enumerate(
                tqdm(
                    dataloader,
                    desc=f"Evaluation on dataloader `{dl_id}`",
                    bar_format="{l_bar}{r_bar}",
                    dynamic_ncols=True,
                    disable=not self.accelerator.is_local_main_process,
                )
            ):
                """
                It is advised against computing metrics within the `evaluation_epoch_end` method for several reasons:
                    1. Most evaluation metrics are inherently sequential and not parallelizable. Hence, computing them in `evaluation_epoch_end` does not offer a speed advantage during evaluation.
                    2. By not aggregating all outputs for metric calculation at the epoch's end, we reduce the risk of memory overflow, which can occur when gathering results across all processes.
                    3. Calculating the metric score during `evaluation_step` allows for earlier detection of any errors in the code.

                Recommendations for metric calculation:
                    1. Perform immediate metric score calculation within the `evaluation_step` method.
                    2. Accumulate the results at this stage.
                    3. If necessary, compute the average or aggregate metric score in the `evaluation_epoch_end` method.
                """
                with torch.no_grad():
                    step_output = self.evaluation_step(batch, batch_idx, dl_id)

                # If `gather_step_output` is True, we will gather the step_output from all processes and return a list of all metric scores.
                if gather_step_output:
                    """Collect the step_output from all processes and return a list of all metric scores.

                    Assume we have two processes:
                    step_output = [
                        {"metric_1": xx, "metric_2": xx, ...},  # process 0
                        {"metric_1": xx, "metric_2": xx, ...},  # process 1
                        {"metric_1": xx, "metric_2": xx, ...},  # process 0
                        {"metric_1": xx, "metric_2": xx, ...},  # process 1
                        ...
                    ]
                    """
                    step_output = self.accelerator.gather_for_metrics(step_output)
                dataloader_output.append(step_output)

            evaluation_output[dl_id] = dataloader_output

        """
        evaluation_output = {
            "dataloader_id_1": [step_output_0, step_output_1, ...],
            "dataloader_id_2": [step_output_0, step_output_1, ...],
            ...
        }
        """
        return evaluation_output

    def _check_improvement(self, score, save_max_score=True):
        """Check if the current model got the best metric score."""
        if save_max_score:
            if score > self.state.best_score:
                return True
            else:
                return False
        else:
            if score < self.state.best_score:
                return True
            else:
                return False

    def training_step(self, batch, batch_idx):
        """Implement a training step (iteration).

        Implement your own training step here. The input batch is from a training dataloader and the output of this
        function can be various. For example, it can be a dict of loss, or a dict of loss and some enhanced audio signals.
        Here is the persuade code for training a model:

        .. code-block:: python
            :emphasize-lines: 6

            for epoch in range(start_epoch, end_epoch):
                self.model.train()

                training_epoch_output = []
                for batch, batch_index in dataloader:
                    loss_dict = training_step(batch, batch_idx)
                    training_epoch_output.append(loss_dict)

                training_epoch_end(training_epoch_output)
                save_checkpoint()
                if some_condition:
                    score = validate()
                    if score > best_score:
                        save_checkpoint(best=True)

        Args:
            batch: a batch of data, which passed from a custom training dataloader.
            batch_idx: the index of the current batch.

        Returns:
            loss_dict: a dict of loss. For example, {"loss_1": loss, "loss_2": loss, ...}
        """
        raise NotImplementedError

    def training_epoch_end(self, training_epoch_output):
        """Implement the logic of the end of a training epoch.

        By default, this function will log the mean loss of each loss item on a training epoch. You can override this.
        When the training epoch ends, this function will be called. The input is a list of the loss dict of each step
        in a training epoch. You may want to log the epoch-level training loss here.

        .. code-block:: python
            for epoch in range(start_epoch, end_epoch):
                self.model.train()

                training_epoch_output = []
                for batch, batch_index in dataloader:
                    loss = training_step(batch, batch_idx)
                    training_epoch_output.append(loss)

                training_epoch_end(training_epoch_output)
                save_checkpoint()
                if some_condition:
                    score = validate()
                    if score > best_score:
                        save_checkpoint(best=True)

        Args:
            training_epoch_output: the output of the training epoch. It may a list of the output of each batch (iteration).
        """
        # Compute mean loss on all loss items on a epoch
        if self.accelerator.is_local_main_process:
            loss_keys = training_epoch_output[0].keys()
            loss_dict = {key: np.mean([step_out[key] for step_out in training_epoch_output]) for key in loss_keys}

            for key, value in loss_dict.items():
                logger.info(f"Loss '{key}' on epoch {self.state.epochs_trained}: {value}")
                self.writer.add_scalar(f"Train_Epoch/{key}", value, self.state.epochs_trained)

            loss_dict["epoch"] = self.state.epochs_trained

            # Append the loss to the loss log
            df = pd.DataFrame(loss_dict, index=[0])
            df.to_csv(self.loss_log_path, mode="a", index=False, header=not self.loss_log_path.exists())

    def evaluation_step(self, batch, batch_idx, dataloader_idx):
        """Implement a evaluation/prediction step.

        This function defines the evaluation step. The input batch is from a eval dataloader.
        Here is the persuade code for validating a model:

        .. code-block:: python
            :emphasize-lines: 4

            evaluation_output = []
            for dataloader_idx, dataloader in dataloaders:
                for batch_index, batch in dataloader:
                    loss_or_data = evaluation_step(batch, batch_idx)
                    evaluation_epoch_output.append(loss_or_data)

            score = evaluation_epoch_end(evaluation_epoch_output)
            return score

        Notes:
            **The evaluation step will be run on all processes.**

            About batch size:
            If your evaluation data have the same length, you may use a batch size larger than 1 to speed up the evaluation.
            For example, if you have 1000 samples in the evaluation set, and you have a batch size of 100, then you will
            have 10 batches in the evaluation set. However, if your data in the evaluation set has a different length, please
            use a batch size of 1. It still works for distributed evaluation. Otherwise, you will get an error.

            About distributed evaluation:
            The output of this function will be gathered across all processes. For example, if you have 4 processes, and
            you have a batch size of 1, then you will have 4 outputs from this function. The output of this function will
            be gathered across all processes. The first dimension of the result is num_processes multiplied by the first
            dimension of the input tensors. **Please make sure the first dimension of the input tensors is the batch size.**
            **The last dimension of the output will be padded to the length of the longest sample in the evaluation set.**
            It means that the output will be a tensor with the shape of [num_processes * batch_size, max_length]. If you
            calculate the metric score on the output, you should do a truncation to remove the padding. Otherwise, if you
            are using a metric that sensitive to the padding, you will get a wrong metric score. It is not easy to
            implement this truncation in the ``evaluation_epoch_end`` function. We recommend you directly calculate the metric
            score in the evaluation_step function. I guess the Accelerate team will implement a automatic truncation in the
            future. https://github.com/huggingface/accelerate/issues/226

        Args:
            batch: a batch of data.
            batch_idx: the index of the batch.
            dataloader_idx: the index of the dataloader.

        Returns:
            output: the output of the batch. It may enhanced audio signals.
        """
        raise NotImplementedError

    def evaluation_epoch_end(self, evaluation_epoch_output):
        """Evaluation epoch end.

        The input `evaluation_epoch_output` will be a list of list. For example, if you have two dataloaders, the `evaluation_epoch_output` will be:

        .. code-block:: python

            evaluation_epoch_output = [
                [dataloader_1_batch_1_output, dataloader_1_batch_2_output, ...],
                [dataloader_2_batch_1_output, dataloader_2_batch_2_output, ...],
                ...,
            ]


        The output of this function should be a metric score, which will be used to determine whether the current model is the best model.

        .. code-block:: python
            :emphasize-lines: 7

            evaluation_output = []
            for dataloader_idx, dataloader in dataloaders:
                for batch_index, batch in dataloader:
                    loss_or_data = evaluation_step(batch, batch_idx)
                    evaluation_epoch_output.append(loss_or_data)

            score = evaluation_epoch_end(evaluation_epoch_output)
            return score

        Args:
            evaluation_epoch_output: the output of the evaluation epoch. It is a list of list.

        Returns:
            score: the metric score of the evaluation epoch.
        """
        raise NotImplementedError



class Metric:
    """Metric class to store the metric name and its value."""

    def __init__(self, name, value) -> None:
        self.name = name
        self.value = value

    def __str__(self):
        return f"{self.name}: {self.value}"

    def __add__(self, other):
        if not isinstance(other, Metric):
            raise ValueError(f"Cannot add Metric with {type(other)}")

        return {self.name: self.value + other.value}

    def __div__(self, other):
        if not isinstance(other, Metric):
            raise ValueError(f"Cannot divide Metric with {type(other)}")

        return {self.name: self.value / other.value}

    def __rdiv__(self, other):
        if not isinstance(other, Metric):
            raise ValueError(f"Cannot divide Metric with {type(other)}")

        return {self.name: other / self.value}


if __name__ == "__main__":
    metric1 = Metric("OVRL", 0.5)
    metric2 = Metric("SIG", 0.6)

    print(metric1 + metric2)
    print(metric1 / metric2)
    print(metric1 + 1)
